import functools
import random
from dataclasses import dataclass
from enum import Enum
from typing import TypeAlias

from fastecdsa import curvemath

from garaga import garaga_rs
from garaga.algebra import (
    BaseField,
    BaseFp2Field,
    Fp2,
    ModuloCircuitElement,
    Polynomial,
    PyFelt,
)
from garaga.hints.io import bigint_split, int_to_u256, int_to_u384

N_LIMBS = 4
BASE = 2**96
STARK = 0x800000000000011000000000000000000000000000000000000000000000001
BN254_ID = 0
BLS12_381_ID = 1
SECP256K1_ID = 2
SECP256R1_ID = 3
ED25519_ID = 4
GRUMPKIN_ID = 5


class ProofSystem(Enum):
    Groth16 = "groth16"
    Risc0Groth16 = "risc0_groth16"
    UltraKeccakHonk = "ultra_keccak_honk"
    UltraStarknetHonk = "ultra_starknet_honk"
    UltraKeccakZKHonk = "ultra_keccak_zk_honk"
    UltraStarknetZKHonk = "ultra_starknet_zk_honk"

    @property
    def supported_curves(self) -> set[int]:
        if self == ProofSystem.Groth16:
            return {BN254_ID, BLS12_381_ID}
        if self == ProofSystem.Risc0Groth16:
            return {BN254_ID}
        if self == ProofSystem.UltraKeccakHonk:
            return {BN254_ID}
        if self == ProofSystem.UltraStarknetHonk:
            return {BN254_ID}
        if self == ProofSystem.UltraKeccakZKHonk:
            return {BN254_ID}
        if self == ProofSystem.UltraStarknetZKHonk:
            return {BN254_ID}
        return set()


class CurveID(Enum):
    BN254 = 0
    BLS12_381 = 1
    SECP256K1 = 2
    SECP256R1 = 3
    ED25519 = 4
    GRUMPKIN = 5

    @staticmethod
    def from_str(s: str) -> "CurveID":
        return CurveID(CurveID.find_value_in_string(s))

    @property
    def p(self) -> int:
        return CURVES[self.value].p

    @staticmethod
    def find_value_in_string(s: str) -> int | None:
        """
        Find the value of the curve ID in the string.
        """
        if s.lower() == "bn128":
            return CurveID.BN254.value
        if s.lower() == "bls12381":
            return CurveID.BLS12_381.value
        for member in CurveID:
            if s.lower() in member.name.lower() or member.name.lower() in s.lower():
                return member.value
        return None

    @staticmethod
    def get_proving_system_curve(
        curve_id: int, proving_system: ProofSystem
    ) -> "CurveID":
        """
        Get the curve ID for proving systems. Only curves supported by the given proving system are valid.
        """
        if curve_id not in CurveID._value2member_map_:
            raise ValueError(
                f"Invalid curve ID: {curve_id}. Supported curves are: {', '.join([f'{c.name} (ID: {c.value})' for c in CurveID])}"
            )
        if curve_id not in proving_system.supported_curves:
            supported_curves = ", ".join(
                f"{CurveID(c).name} (ID: {c})" for c in proving_system.supported_curves
            )
            raise ValueError(
                f"Invalid curve ID for {proving_system.name}. Supported curves are: {supported_curves}"
            )
        return CurveID(curve_id)


@dataclass
class SWUParams:
    A: int
    B: int
    Z: int


@dataclass(slots=True, frozen=True)
class WeierstrassCurve:
    cairo_zero_namespace_name: str
    id: int
    p: int
    n: int  # order
    h: int  # cofactor
    a: int  # y^2 = x^3 + ax + b
    b: int
    fp_generator: int  # A generator of the field of the curve. To verify it, use is_generator function.
    Gx: int  # x-coordinate of the generator point
    Gy: int  # y-coordinate of the generator point
    swu_params: SWUParams
    eigen_value: int | None  # Endomorphism eigenvalue
    third_root_of_unity: int | None  # Endomorphism image

    def is_endomorphism_available(self) -> bool:
        return self.eigen_value is not None and self.third_root_of_unity is not None

    def to_cairo_zero(self) -> str:
        code = f"namespace {self.cairo_zero_namespace_name} {{\n"
        code += f"const CURVE_ID = {self.id};\n"
        p = bigint_split(self.p, N_LIMBS, BASE)
        n = bigint_split(self.n, N_LIMBS, BASE)
        a = bigint_split(self.a, N_LIMBS, BASE)
        b = bigint_split(self.b, N_LIMBS, BASE)
        g = bigint_split(self.fp_generator, N_LIMBS, BASE)
        min_one = bigint_split(-1 % self.p, N_LIMBS, BASE)
        for i, limb in enumerate(p):
            code += f"const P{i} = {hex(limb)};\n"
        for i, limb in enumerate(n):
            code += f"const N{i} = {hex(limb)};\n"
        for i, limb in enumerate(a):
            code += f"const A{i} = {hex(limb)};\n"
        for i, limb in enumerate(b):
            code += f"const B{i} = {hex(limb)};\n"
        for i, limb in enumerate(g):
            code += f"const G{i} = {hex(limb)};\n"
        for i, limb in enumerate(min_one):
            code += f"const MIN_ONE_D{i} = {hex(limb)};\n"
        code += "}\n"
        return code

    def to_cairo_one(self) -> str:
        code = f"const {self.cairo_zero_namespace_name.upper()}:Curve = \n"
        code += "Curve {\n"
        code += f"p:{int_to_u384(self.p)},\n"
        code += f"n:{int_to_u256(self.n)},\n"
        code += f"a:{int_to_u384(self.a)},\n"
        code += f"b:{int_to_u384(self.b)},\n"
        code += f"g:{int_to_u384(self.fp_generator)},\n"
        code += f"min_one:{int_to_u384(-1%self.p)},\n"
        code += f"G: {G1Point(self.Gx, self.Gy, CurveID(self.id)).to_cairo_1()},\n"
        code += "};\n"
        return code


@dataclass(slots=True, frozen=True)
class TwistedEdwardsCurve(WeierstrassCurve):
    """
    Twisted Edwards curve is a curve of the form ax^2 + y^2 = 1 + dx^2y^2.
    Equivalent to a Weierstrass curve of the form y^2 = x^3 + ax + b.
    Automatic conversion of parameters is done in the constructor.
    """

    a_twisted: int  # Twisted Edwards curve parameter
    d_twisted: int  # Twisted Edwards curve parameter

    def __init__(
        self,
        cairo_zero_namespace_name: str,
        id: int,
        p: int,
        n: int,
        h: int,
        d_twisted: int,
        a_twisted: int,
        fp_generator: int,
        Gx: int,
        Gy: int,
        swu_params: SWUParams,
        eigen_value: int | None = None,
        third_root_of_unity: int | None = None,
    ):
        assert a_twisted != 0 and d_twisted != 0 and a_twisted != d_twisted
        # Set attributes
        object.__setattr__(self, "d_twisted", d_twisted)
        object.__setattr__(self, "a_twisted", a_twisted)
        object.__setattr__(self, "p", p)
        object.__setattr__(self, "swu_params", swu_params)
        object.__setattr__(self, "eigen_value", eigen_value)
        object.__setattr__(self, "third_root_of_unity", third_root_of_unity)
        # Calculate Weierstrass parameters
        a = (
            -1
            * pow(48, -1, p)
            * (a_twisted**2 + 14 * a_twisted * d_twisted + d_twisted**2)
            % p
        )
        b = (
            pow(864, -1, p)
            * (a_twisted + d_twisted)
            * (-(a_twisted**2) + 34 * a_twisted * d_twisted - d_twisted**2)
            % p
        )

        WeierstrassCurve.__init__(
            self,
            cairo_zero_namespace_name,
            id,
            p,
            n,
            h,
            a,
            b,
            fp_generator,
            *(self.to_weierstrass(Gx, Gy)),
            swu_params,
            eigen_value,
            third_root_of_unity,
        )

    def to_weierstrass(self, x_twisted, y_twisted):
        a = self.a_twisted
        d = self.d_twisted
        return (
            (5 * a + a * y_twisted - 5 * d * y_twisted - d)
            * pow(12 - 12 * y_twisted, -1, self.p)
            % self.p,
            (a + a * y_twisted - d * y_twisted - d)
            * pow(4 * x_twisted - 4 * x_twisted * y_twisted, -1, self.p)
            % self.p,
        )

    def to_twistededwards(self, x_weirstrass: int, y_weirstrass: int):
        a = self.a_twisted
        d = self.d_twisted
        y = (
            (5 * a - 12 * x_weirstrass - d)
            * pow(-12 * x_weirstrass - a + 5 * d, -1, self.p)
            % self.p
        )
        x = (
            (a + a * y - d * y - d)
            * pow(4 * y_weirstrass - 4 * y_weirstrass * y, -1, self.p)
            % self.p
        )
        return (x, y)


@dataclass(slots=True, frozen=True)
class PairingCurve(WeierstrassCurve):
    x: int  # curve x parameter
    irreducible_polys: dict[int, list[int]]
    nr_a0: int  # E2 non residue
    nr_a1: int
    b20: int
    b21: int  # E2: b is (b20, b21)
    loop_counter: list[int]
    line_function_sparsity: list[
        int
    ]  # 0: ==0, 1: !=0, 2: ==1.. L(x) = Σ(sparsity[i] * coeff[i] * x^i )
    final_exp_cofactor: int
    G2x: tuple[int, int]
    G2y: tuple[int, int]


def NAF(x):
    if x == 0:
        return []
    z = 0 if x % 2 == 0 else 2 - (x % 4)
    return NAF((x - z) // 2) + [z]


def jy00(value: int) -> list[int]:
    """
    This is a minimum-Hamming-Weight left-to-right recoding.
    It outputs signed {-1, 0, 1} bits from MSB to LSB
    with minimal Hamming Weight to minimize operations
    in Miller Loops and vartime scalar multiplications

    - Optimal Left-to-Right Binary Signed-Digit Recoding
      Joye, Yen, 2000
      https://marcjoye.github.io/papers/JY00sd2r.pdf
    """

    def bit(value, index):
        return (value >> index) & 1

    bi, bi1, ri, ri1, ri2 = 0, 0, 0, 0, 0
    bits = value.bit_length()
    recoded = []

    for i in range(bits, -1, -1):
        if i == bits:
            ri1, ri2 = bit(value, bits - 1), bit(value, bits - 2)
        else:
            bi, ri = bi1, ri1
            ri1, ri2 = ri2, bit(value, i - 2) if i >= 2 else 0

        bi1 = (bi + ri1 + ri2) >> 1
        recoded.append(-2 * bi + ri + bi1)

    return recoded


GARAGA_RS_SUPPORTED_CURVES = {BN254_ID, BLS12_381_ID}

Curve: TypeAlias = WeierstrassCurve

CURVES: dict[int, WeierstrassCurve] = {
    BN254_ID: PairingCurve(
        cairo_zero_namespace_name="bn",
        id=BN254_ID,
        p=0x30644E72E131A029B85045B68181585D97816A916871CA8D3C208C16D87CFD47,
        n=0x30644E72E131A029B85045B68181585D2833E84879B9709143E1F593F0000001,
        x=0x44E992B44A6909F1,
        h=1,
        irreducible_polys={
            6: [82, 0, 0, -18, 0, 0, 1],
            12: [82, 0, 0, 0, 0, 0, -18, 0, 0, 0, 0, 0, 1],
        },
        nr_a0=9,
        nr_a1=1,
        a=0,
        b=3,
        b20=0x2B149D40CEB8AAAE81BE18991BE06AC3B5B4C5E559DBEFA33267E6DC24A138E5,
        b21=0x9713B03AF0FED4CD2CAFADEED8FDF4A74FA084E52D1852E4A2BD0685C315D2,
        loop_counter=jy00(6 * 0x44E992B44A6909F1 + 2)[::-1],
        line_function_sparsity=[
            2,
            1,
            0,
            1,
            0,
            0,
            0,
            1,
            0,
            1,
            0,
            0,
        ],
        final_exp_cofactor=1469306990098747947464455738335385361638823152381947992820,  # cofactor = 2 * x0 * (6 * x0**2 + 3 * x0 + 1)
        fp_generator=3,
        Gx=0x1,
        Gy=0x2,
        G2x=(
            0x1800DEEF121F1E76426A00665E5C4479674322D4F75EDADD46DEBD5CD992F6ED,
            0x198E9393920D483A7260BFB731FB5D25F1AA493335A9E71297E485B7AEF312C2,
        ),
        G2y=(
            0x12C85EA5DB8C6DEB4AAB71808DCB408FE3D1E7690C43D37B4CE6CC0166FA7DAA,
            0x90689D0585FF075EC9E99AD690C3395BC4B313370B38EF355ACDADCD122975B,
        ),
        swu_params=None,
        eigen_value=0xB3C4D79D41A917585BFC41088D8DAAA78B17EA66B99C90DD,
        third_root_of_unity=0x59E26BCEA0D48BACD4F263F1ACDB5C4F5763473177FFFFFE,
    ),
    BLS12_381_ID: PairingCurve(
        cairo_zero_namespace_name="bls",
        id=BLS12_381_ID,
        p=0x1A0111EA397FE69A4B1BA7B6434BACD764774B84F38512BF6730D2A0F6B0F6241EABFFFEB153FFFFB9FEFFFFFFFFAAAB,
        n=0x73EDA753299D7D483339D80809A1D80553BDA402FFFE5BFEFFFFFFFF00000001,
        x=-0xD201000000010000,
        h=0x396C8C005555E1568C00AAAB0000AAAB,
        irreducible_polys={
            6: [2, 0, 0, -2, 0, 0, 1],
            12: [2, 0, 0, 0, 0, 0, -2, 0, 0, 0, 0, 0, 1],
        },
        nr_a0=1,
        nr_a1=1,
        a=0,
        b=4,
        b20=4,
        b21=4,
        loop_counter=[int(x) for x in bin(0xD201000000010000)[2:][::-1]],
        line_function_sparsity=[
            1,
            0,
            1,
            2,
            0,
            0,
            1,
            0,
            1,
            0,
            0,
            0,
        ],
        final_exp_cofactor=3,
        fp_generator=3,
        Gx=0x17F1D3A73197D7942695638C4FA9AC0FC3688C4F9774B905A14E3A3F171BAC586C55E83FF97A1AEFFB3AF00ADB22C6BB,
        Gy=0x08B3F481E3AAA0F1A09E30ED741D8AE4FCF5E095D5D00AF600DB18CB2C04B3EDD03CC744A2888AE40CAA232946C5E7E1,
        G2x=(
            0x24AA2B2F08F0A91260805272DC51051C6E47AD4FA403B02B4510B647AE3D1770BAC0326A805BBEFD48056C8C121BDB8,
            0x13E02B6052719F607DACD3A088274F65596BD0D09920B61AB5DA61BBDC7F5049334CF11213945D57E5AC7D055D042B7E,
        ),
        G2y=(
            0xCE5D527727D6E118CC9CDC6DA2E351AADFD9BAA8CBDD3A76D429A695160D12C923AC9CC3BACA289E193548608B82801,
            0x606C4A02EA734CC32ACD2B02BC28B99CB3E287E85A763AF267492AB572E99AB3F370D275CEC1DA1AAA9075FF05F79BE,
        ),
        swu_params=SWUParams(
            A=0x144698A3B8E9433D693A02C96D4982B0EA985383EE66A8D8E8981AEFD881AC98936F8DA0E0F97F5CF428082D584C1D,
            B=0x12E2908D11688030018B12E8753EEE3B2016C1F0F24F4070A0B9C14FCEF35EF55A23215A316CEAA5D1CC48E98E172BE0,
            Z=11,
        ),
        eigen_value=0xAC45A4010001A40200000000FFFFFFFF,
        third_root_of_unity=0x1A0111EA397FE699EC02408663D4DE85AA0D857D89759AD4897D29650FB85F9B409427EB4F49FFFD8BFD00000000AAAC,
    ),
    SECP256K1_ID: WeierstrassCurve(
        cairo_zero_namespace_name="secp256k1",
        id=SECP256K1_ID,
        p=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEFFFFFC2F,
        n=0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEBAAEDCE6AF48A03BBFD25E8CD0364141,
        h=1,
        a=0,
        b=7,
        fp_generator=3,
        Gx=0x79BE667EF9DCBBAC55A06295CE870B07029BFCDB2DCE28D959F2815B16F81798,
        Gy=0x483ADA7726A3C4655DA4FBFC0E1108A8FD17B448A68554199C47D08FFB10D4B8,
        swu_params=None,
        eigen_value=0x5363AD4CC05C30E0A5261C028812645A122E22EA20816678DF02967C1B23BD72,
        third_root_of_unity=0x7AE96A2B657C07106E64479EAC3434E99CF0497512F58995C1396C28719501EE,
    ),
    SECP256R1_ID: WeierstrassCurve(
        cairo_zero_namespace_name="secp256r1",
        id=SECP256R1_ID,
        p=0xFFFFFFFF00000001000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFF,
        n=0xFFFFFFFF00000000FFFFFFFFFFFFFFFFBCE6FAADA7179E84F3B9CAC2FC632551,
        h=1,
        a=0xFFFFFFFF00000001000000000000000000000000FFFFFFFFFFFFFFFFFFFFFFFC,
        b=0x5AC635D8AA3A93E7B3EBBD55769886BC651D06B0CC53B0F63BCE3C3E27D2604B,
        fp_generator=6,
        Gx=0x6B17D1F2E12C4247F8BCE6E563A440F277037D812DEB33A0F4A13945D898C296,
        Gy=0x4FE342E2FE1A7F9B8EE7EB4A7C0F9E162BCE33576B315ECECBB6406837BF51F5,
        swu_params=None,
        eigen_value=None,
        third_root_of_unity=None,
    ),
    ED25519_ID: TwistedEdwardsCurve(
        cairo_zero_namespace_name="ED25519",  # See https://neuromancer.sk/std/other/Ed25519
        id=ED25519_ID,
        p=0x7FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFED,
        n=0x1000000000000000000000000000000014DEF9DEA2F79CD65812631A5CF5D3ED,
        h=8,
        d_twisted=0x52036CEE2B6FFE738CC740797779E89800700A4D4141D8AB75EB4DCA135978A3,
        a_twisted=0x7FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEC,
        fp_generator=6,
        Gx=0x216936D3CD6E53FEC0A4E231FDD6DC5C692CC7609525A7B2C9562D608F25D51A,
        Gy=0x6666666666666666666666666666666666666666666666666666666666666658,
        swu_params=None,
        eigen_value=None,
        third_root_of_unity=None,
    ),
    GRUMPKIN_ID: WeierstrassCurve(
        cairo_zero_namespace_name="grumpkin",
        id=GRUMPKIN_ID,
        p=0x30644E72E131A029B85045B68181585D2833E84879B9709143E1F593F0000001,
        n=0x30644E72E131A029B85045B68181585D97816A916871CA8D3C208C16D87CFD47,
        h=1,
        a=0,
        b=-17 % 0x30644E72E131A029B85045B68181585D2833E84879B9709143E1F593F0000001,
        fp_generator=5,
        Gx=0x1,
        Gy=0x2CF135E7506A45D632D270D45F1181294833FC48D823F272C,
        swu_params=None,
        eigen_value=None,
        third_root_of_unity=None,
    ),
}


def is_generator(g: int, p: int) -> bool:
    """
    Checks if a given integer g is a generator of the multiplicative group of integers modulo p.

    Parameters:
    g (int): The integer to check.
    p (int): The prime number defining the finite field Fp.

    Returns:
    bool: True if g is a generator, False otherwise.

    Raises:
    ValueError: If p is not a prime number.

    Detailed Explanation:
    1. The order of an element g in the multiplicative group Fp* must divide p-1.
       - The multiplicative group Fp* consists of all non-zero elements of the finite field Fp.
       - The group has p-1 elements because there are p-1 non-zero elements in Fp.
       - By Lagrange's theorem, the order of any element in a finite group must divide the order of the group.
       - Therefore, the order of any element g in Fp* must divide p-1.
    2. If g is a generator, its order must be exactly p-1.
    3. To verify that g has order p-1, we need to ensure that g does not have any smaller order that divides p-1.
    4. The prime factorization of p-1 gives us all the possible divisors of p-1.
    5. By checking g^((p-1)/q) for each prime factor q of p-1, we ensure that g does not have an order that is a proper divisor of p-1.
    6. If g passes all these checks, it must have the full order p-1, making it a generator.
    7. This method is efficient because it reduces the problem to checking a few exponentiations rather than testing all possible orders.

    Example:
    - Let p = 7, then p-1 = 6.
    - The prime factors of 6 are 2 and 3.
    - To check if g is a generator, we test:
      - g^(6/2) = g^3
      - g^(6/3) = g^2
    - If neither g^3 ≡ 1 (mod 7) nor g^2 ≡ 1 (mod 7), then g is a generator.
    - This ensures that g has the full order 6, covering all elements of F7*.
    """
    import sympy

    # Ensure that p is a prime number
    if not sympy.isprime(p):
        raise ValueError("p must be a prime number.")

    # Calculate p - 1, which is the order of the multiplicative group of integers modulo p
    p_minus_1 = p - 1

    # Factorize p - 1 to get its prime factors and their exponents
    # This is necessary because we will use these factors to test if g is a generator
    factors = sympy.factorint(p_minus_1)

    # Check if g is a generator
    for q in factors.keys():
        # For each prime factor q of p - 1, check if g^((p-1)/q) ≡ 1 (mod p)
        # If g^((p-1)/q) ≡ 1 (mod p), then g is not a generator because it does not have the full order p-1
        # This is based on the fact that if g is a generator, its order must be exactly p-1
        # If g^((p-1)/q) ≡ 1 (mod p), it means the order of g divides (p-1)/q, which is less than p-1
        if pow(g, p_minus_1 // q, p) == 1:
            return False

    # If g^((p-1)/q) ≠ 1 (mod p) for all prime factors q, then g is a generator
    # This means g has the full order p-1 and can generate all elements of the multiplicative group
    return True


def get_base_field(
    curve_id: int | CurveID, field_type: type[PyFelt] | type[Fp2] = PyFelt
) -> BaseField | BaseFp2Field:
    """
    Returns the base field for a given elliptic curve.

    Parameters:
    curve_id (int | CurveID): The ID of the elliptic curve.
    field_type (PyFelt | Fp2): The type of the field (default is PyFelt).

    Returns:
    BaseField | BaseFp2Field: The base field corresponding to the curve and field type.

    Raises:
    ValueError: If the field_type is invalid.
    """
    if isinstance(curve_id, CurveID):
        curve_id = curve_id.value

    curve = CURVES[curve_id]
    if field_type == PyFelt:
        return BaseField(curve.p)
    elif field_type == Fp2:
        assert isinstance(curve, PairingCurve)
        return BaseFp2Field(curve.p)
    else:
        raise ValueError(f"Invalid field type: {field_type}. Expected PyFelt or Fp2.")


def get_irreducible_poly(curve_id: int | CurveID, extension_degree: int) -> Polynomial:
    if isinstance(curve_id, CurveID):
        curve_id = curve_id.value
    curve = CURVES[curve_id]
    assert isinstance(curve, PairingCurve)
    field = get_base_field(curve_id)
    return Polynomial(
        coefficients=[field(x) for x in curve.irreducible_polys[extension_degree]]
    )


@dataclass(slots=True)
class G1Point:
    """
    Represents a point on G1, the group of rational points on an elliptic curve over the base field.

    Attributes:
        x (int): The x-coordinate of the point.
        y (int): The y-coordinate of the point.
        curve_id (CurveID): The identifier of the elliptic curve.
    """

    x: int
    y: int
    curve_id: CurveID
    iso_point: bool = False

    def __repr__(self) -> str:
        return f"G1Point({hex(self.x)}, {hex(self.y)}) on {self.curve_id.value}"

    def __str__(self) -> str:
        return f"G1Point({self.x}, {self.y}) on curve {self.curve_id}"

    def __hash__(self):
        return hash((self.x, self.y, self.curve_id, self.iso_point))

    def __eq__(self, other: object) -> bool:
        """
        Checks if two G1Point instances are equal.

        Args:
            other (G1Point): The other point to compare.

        Returns:
            bool: True if the points are equal, False otherwise.
        """
        if not isinstance(other, G1Point):
            raise ValueError(f"Cannot compare G1Point with {type(other)}")
        return (
            self.x == other.x
            and self.y == other.y
            and self.curve_id.value == other.curve_id.value
            and self.iso_point == other.iso_point
        )

    def __post_init__(self):
        """
        Post-initialization checks to ensure the point is valid.
        """
        if self.is_infinity():
            return
        if not self.is_on_curve():
            raise ValueError(f"Point {self} is not on the curve {self.curve_id}")

    @staticmethod
    def infinity(curve_id: CurveID) -> "G1Point":
        """
        Returns the point at infinity for the given curve.

        Args:
            curve_id (CurveID): The identifier of the elliptic curve.

        Returns:
            G1Point: The point at infinity.
        """
        return G1Point(0, 0, curve_id)

    def is_infinity(self) -> bool:
        """
        Checks if the point is the point at infinity.

        Returns:
            bool: True if the point is at infinity, False otherwise.
        """
        return self.x == 0 and self.y == 0

    def to_cairo_1(self, as_hex: bool = True) -> str:
        """
        Converts the point to a Cairo 1 representation.

        Returns:
            str: The Cairo 1 representation of the point.
        """
        return f"G1Point{{x: {int_to_u384(self.x, as_hex)}, y: {int_to_u384(self.y, as_hex)}}}"

    @staticmethod
    def gen_random_point_not_in_subgroup(
        curve_id: CurveID, force_gen: bool = False
    ) -> "G1Point":
        """
        Generates a random point not in the prime order subgroup.

        Args:
            curve_id (CurveID): The identifier of the elliptic curve.
            force_gen (bool): Force generation even if the cofactor is 1.

        Returns:
            G1Point: A random point not in the prime order subgroup.

        Raises:
            ValueError: If the cofactor is 1 and force_gen is False.
        """
        curve_idx = curve_id.value
        if CURVES[curve_idx].h == 1:
            if force_gen:
                return G1Point.gen_random_point(curve_id)
            else:
                raise ValueError(
                    "Cofactor is 1, cannot generate a point not in the subgroup"
                )
        else:
            field: BaseField = get_base_field(curve_idx)
            while True:
                x = field.random()
                y2 = x**3 + CURVES[curve_idx].a * x + CURVES[curve_idx].b
                try:
                    tentative_point = G1Point(x.value, y2.sqrt().value, curve_id)
                except ValueError:
                    continue
                if not tentative_point.is_in_prime_order_subgroup():
                    return tentative_point

    def is_in_prime_order_subgroup(self) -> bool:
        """
        Checks if the point is in the prime order subgroup.

        Returns:
            bool: True if the point is in the prime order subgroup, False otherwise.
        """
        return self.scalar_mul(CURVES[self.curve_id.value].n).is_infinity()

    def is_on_curve(self) -> bool:
        """
        Checks if the point is on the curve using the curve equation y^2 = x^3 + ax + b.

        Returns:
            bool: True if the point is on the curve, False otherwise.
        """
        if self.iso_point:
            a = CURVES[self.curve_id.value].swu_params.A
            b = CURVES[self.curve_id.value].swu_params.B
        else:
            a = CURVES[self.curve_id.value].a
            b = CURVES[self.curve_id.value].b

        p = CURVES[self.curve_id.value].p
        lhs = self.y**2 % p
        rhs = (self.x**3 + a * self.x + b) % p
        return lhs == rhs

    @staticmethod
    def gen_random_point(curve_id: CurveID) -> "G1Point":
        """
        Generates a random point on a given curve.

        Args:
            curve_id (CurveID): The identifier of the elliptic curve.

        Returns:
            G1Point: A random point on the curve.
        """
        scalar = random.randint(1, CURVES[curve_id.value].n - 1)

        return G1Point.get_nG(curve_id, scalar)

    @staticmethod
    def get_nG(curve_id: CurveID, n: int) -> "G1Point":
        """
        Returns the scalar multiplication of the generator point on a given curve by the scalar n.

        Args:
            curve_id (CurveID): The identifier of the elliptic curve.
            n (int): The scalar to multiply by.

        Returns:
            G1Point: The resulting point after scalar multiplication.

        Raises:
            AssertionError: If n is not less than the order of the curve.
        """
        assert (
            n < CURVES[curve_id.value].n
        ), "n must be less than the order of the curve"

        gen = G1Point(CURVES[curve_id.value].Gx, CURVES[curve_id.value].Gy, curve_id)
        return gen.scalar_mul(n)

    @staticmethod
    def msm(points: list["G1Point"], scalars: list[int]) -> "G1Point":
        """
        Performs multi-scalar multiplication (MSM) on a list of points and scalars.

        Args:
            points (list[G1Point]): The list of points.
            scalars (list[int]): The list of scalars.

        Returns:
            G1Point: The resulting point after MSM.
        """
        assert len(points) == len(
            scalars
        ), f"Points and scalar length mismatch: {len(points)} points and {len(scalars)} scalars"
        muls = [P.scalar_mul(s) for P, s in zip(points, scalars)]
        scalar_mul = functools.reduce(lambda acc, p: acc.add(p), muls)
        return scalar_mul

    def scalar_mul(self, scalar: int) -> "G1Point":
        """
        Performs scalar multiplication on the point.

        Args:
            scalar (int): The scalar to multiply by. abs(scalar) should be less than the order of the curve.

        Returns:
            G1Point: The resulting point after scalar multiplication.
        """
        if self.is_infinity():
            return self
        if scalar == 0:
            return G1Point(0, 0, self.curve_id, self.iso_point)
        if self.iso_point:
            a = CURVES[self.curve_id.value].swu_params.A
            b = CURVES[self.curve_id.value].swu_params.B
        else:
            a = CURVES[self.curve_id.value].a
            b = CURVES[self.curve_id.value].b
        # Fastecdsa C binding.
        x, y = curvemath.mul(
            str(self.x),
            str(self.y),
            str(abs(scalar)),
            str(CURVES[self.curve_id.value].p),
            str(a),
            str(b),
            str(CURVES[self.curve_id.value].n),
            str(CURVES[self.curve_id.value].Gx),
            str(CURVES[self.curve_id.value].Gy),
        )
        # Fastecdsa already returns (0, 0) for the identity element.
        if scalar < 0:
            return -G1Point(int(x), int(y), self.curve_id, self.iso_point)
        else:
            return G1Point(int(x), int(y), self.curve_id, self.iso_point)

    def add(self, other: "G1Point") -> "G1Point":
        """
        Adds two points on the elliptic curve.

        Args:
            other (G1Point): The other point to add.

        Returns:
            G1Point: The resulting point after addition.

        Raises:
            ValueError: If the points are not on the same curve.
        """
        if self.is_infinity():
            return other
        if other.is_infinity():
            return self
        if self.curve_id != other.curve_id:
            raise ValueError("Points are not on the same curve")
        if self.iso_point != other.iso_point:
            raise ValueError("Points are not on the same curve")
        if self.iso_point:
            a = CURVES[self.curve_id.value].swu_params.A
            b = CURVES[self.curve_id.value].swu_params.B
        else:
            a = CURVES[self.curve_id.value].a
            b = CURVES[self.curve_id.value].b

        x, y = curvemath.add(
            str(self.x),
            str(self.y),
            str(other.x),
            str(other.y),
            str(CURVES[self.curve_id.value].p),
            str(a),
            str(b),
            str(CURVES[self.curve_id.value].n),
            str(CURVES[self.curve_id.value].Gx),
            str(CURVES[self.curve_id.value].Gy),
        )
        # NB : Fastecdsa returns (0, 0) for the identity element.
        return G1Point(int(x), int(y), self.curve_id, self.iso_point)

    def __neg__(self) -> "G1Point":
        """
        Negates the point on the elliptic curve.

        Returns:
            G1Point: The negated point.
        """
        return G1Point(
            self.x,
            -self.y % CURVES[self.curve_id.value].p,
            self.curve_id,
            self.iso_point,
        )

    def to_pyfelt_list(self) -> list[PyFelt]:
        field = get_base_field(self.curve_id.value)
        return [field(self.x), field(self.y)]

    def serialize_to_cairo(self, name: str, raw: bool = False) -> str:
        import garaga.modulo_circuit_structs as structs

        return structs.G1PointCircuit(name=name, elmts=self.to_pyfelt_list()).serialize(
            raw
        )


@dataclass(frozen=True)
class G2Point:
    """
    Represents a point on G2, the group of rational points on an elliptic curve over an extension field.
    """

    x: tuple[int, int]
    y: tuple[int, int]
    curve_id: CurveID

    def __repr__(self):
        return f"G2Point({hex(self.x[0])}, {hex(self.x[1])}, {hex(self.y[0])}, {hex(self.y[1])}, {self.curve_id})"

    def __post_init__(self):
        assert isinstance(CURVES[self.curve_id.value], PairingCurve)
        if self.is_infinity():
            return
        if not self.is_on_curve():
            raise ValueError(f"G2 Point is not on the curve {self.curve_id}")

    @staticmethod
    def infinity(curve_id: CurveID) -> "G2Point":
        return G2Point((0, 0), (0, 0), curve_id)

    def __eq__(self, other: "G2Point") -> bool:
        return (
            self.x[0] == other.x[0]
            and self.x[1] == other.x[1]
            and self.y[0] == other.y[0]
            and self.y[1] == other.y[1]
            and self.curve_id == other.curve_id
        )

    def is_infinity(self) -> bool:
        return self.x == (0, 0) and self.y == (0, 0)

    def is_on_curve(self) -> bool:
        """
        Check if the point is on the curve using the curve equation y^2 = x^3 + ax + b in the extension field.
        """
        from garaga.hints.tower_backup import E2

        curve = CURVES[self.curve_id.value]
        assert isinstance(curve, PairingCurve)
        a = curve.a
        p = curve.p
        b = E2(curve.b20, curve.b21, p)
        y = E2(*self.y, p)
        x = E2(*self.x, p)
        return y**2 == x**3 + a * x + b

    @staticmethod
    def gen_random_point(curve_id: CurveID) -> "G2Point":
        """
        Generates a random point on a given curve.
        """
        curve = CURVES[curve_id.value]
        assert isinstance(curve, PairingCurve)
        scalar = random.randint(1, curve.n - 1)
        a = (curve.G2x[0], curve.G2x[1], curve.G2y[0], curve.G2y[1])
        b = garaga_rs.g2_scalar_mul(curve_id.value, a, scalar)
        return G2Point((b[0], b[1]), (b[2], b[3]), curve_id)

    @staticmethod
    def get_nG(curve_id: CurveID, n: int) -> "G2Point":
        """
        Returns the scalar multiplication of the generator point on a given curve by the scalar n.
        """
        assert (
            n < CURVES[curve_id.value].n
        ), "n must be less than the order of the curve"

        if curve_id.value in GARAGA_RS_SUPPORTED_CURVES:
            curve = CURVES[curve_id.value]
            a = (curve.G2x[0], curve.G2x[1], curve.G2y[0], curve.G2y[1])
            b = garaga_rs.g2_scalar_mul(curve_id.value, a, n)
            return G2Point((b[0], b[1]), (b[2], b[3]), curve_id)
        else:
            raise NotImplementedError(
                "G2Point.get_nG is not implemented for this curve"
            )

    def scalar_mul(self, scalar: int) -> "G2Point":
        if self.is_infinity():
            return self
        if scalar == 0:
            return G2Point((0, 0), (0, 0), self.curve_id)
        if scalar < 0:
            return -self.scalar_mul(-scalar)
        if self.curve_id.value in GARAGA_RS_SUPPORTED_CURVES:
            a = (self.x[0], self.x[1], self.y[0], self.y[1])
            b = garaga_rs.g2_scalar_mul(self.curve_id.value, a, scalar)
            return G2Point((b[0], b[1]), (b[2], b[3]), self.curve_id)
        else:
            raise NotImplementedError(
                "G2Point.scalar_mul is not implemented for this curve"
            )

    def add(self, other: "G2Point") -> "G2Point":
        if self.is_infinity():
            return other
        if other.is_infinity():
            return self
        if self.curve_id != other.curve_id:
            raise ValueError("Points are not on the same curve")
        if self.curve_id.value in GARAGA_RS_SUPPORTED_CURVES:
            a = (self.x[0], self.x[1], self.y[0], self.y[1])
            b = (other.x[0], other.x[1], other.y[0], other.y[1])
            c = garaga_rs.g2_add(self.curve_id.value, a, b)
            return G2Point((c[0], c[1]), (c[2], c[3]), self.curve_id)
        else:
            raise NotImplementedError("G2Point.add is not implemented for this curve")

    def __neg__(self) -> "G2Point":
        p = CURVES[self.curve_id.value].p
        return G2Point(
            (self.x[0], self.x[1]), (-self.y[0] % p, -self.y[1] % p), self.curve_id
        )

    @staticmethod
    def msm(points: list["G2Point"], scalars: list[int]) -> "G2Point":
        assert all(isinstance(p, G2Point) for p in points)
        assert len(points) == len(scalars)
        muls = [P.scalar_mul(s) for P, s in zip(points, scalars)]
        scalar_mul = functools.reduce(lambda acc, p: acc.add(p), muls)
        return scalar_mul

    def to_pyfelt_list(self) -> list[PyFelt]:
        field = get_base_field(self.curve_id.value)
        return [field(x) for x in self.x + self.y]

    def serialize_to_cairo(self, name: str, raw: bool = False) -> str:
        import garaga.modulo_circuit_structs as structs

        return structs.G2PointCircuit(name=name, elmts=self.to_pyfelt_list()).serialize(
            raw
        )


@dataclass(slots=True)
class G1G2Pair:
    p: G1Point
    q: G2Point
    curve_id: CurveID = None

    def __hash__(self):
        return hash((self.p, self.q, self.curve_id))

    def __post_init__(self):
        if self.p.curve_id != self.q.curve_id:
            raise ValueError("Points are not on the same curve")
        self.curve_id = self.p.curve_id

    def to_pyfelt_list(self) -> list[PyFelt]:
        field = get_base_field(self.curve_id.value)
        return [
            field(x)
            for x in [
                self.p.x,
                self.p.y,
                self.q.x[0],
                self.q.x[1],
                self.q.y[0],
                self.q.y[1],
            ]
        ]

    @staticmethod
    def pair(pairs: list["G1G2Pair"], curve_id: CurveID = None) -> "E12":
        from garaga.hints.tower_backup import E12  # avoids cycle

        if curve_id is None:
            if len(pairs) == 0:
                raise ValueError("Unspecified curve")
            curve_id = pairs[0].curve_id
        if curve_id.value in GARAGA_RS_SUPPORTED_CURVES:
            args = []
            for pair in pairs:
                if pair.curve_id != curve_id:
                    raise ValueError("Pairs are not on the same curve")
                args.append(pair.p.x)
                args.append(pair.p.y)
                args.append(pair.q.x[0])
                args.append(pair.q.x[1])
                args.append(pair.q.y[0])
                args.append(pair.q.y[1])
            res = garaga_rs.multi_pairing(curve_id.value, args)
            return E12(res, curve_id.value)
        else:
            raise NotImplementedError("G1G2Pair.pair is not implemented for this curve")

    @staticmethod
    def miller(pairs: list["G1G2Pair"], curve_id: CurveID = None):
        from garaga.hints.tower_backup import E12  # avoids cycle

        if curve_id is None:
            if len(pairs) == 0:
                raise ValueError("Unspecified curve")
            curve_id = pairs[0].curve_id
        if curve_id.value in GARAGA_RS_SUPPORTED_CURVES:
            args = []
            for pair in pairs:
                if pair.curve_id != curve_id:
                    raise ValueError("Pairs are not on the same curve")
                args.append(pair.p.x)
                args.append(pair.p.y)
                args.append(pair.q.x[0])
                args.append(pair.q.x[1])
                args.append(pair.q.y[0])
                args.append(pair.q.y[1])
            res = garaga_rs.multi_miller_loop(curve_id.value, args)
            return E12(res, curve_id.value)
        else:
            raise NotImplementedError(
                "G1G2Pair.miller is not implemented for this curve"
            )


# v^6 - 18v^3 + 82
# w^12 - 18w^6 + 82
# v^6 - 2v^3 + 2
# w^12 - 2w^6 + 2


def get_sparsity(X: list[PyFelt | ModuloCircuitElement]) -> list[int]:
    """
    Determines the sparsity of polynomial coefficients.

    This function evaluates a list of polynomial coefficients (`X`) and categorizes each based on its value:
    - 2: The coefficient is 1, indicating a direct representation of a polynomial power.
    - 1: The coefficient is non-zero but not 1, indicating a contributing non-zero coefficient.
    - 0: The coefficient is 0, indicating no contribution to the polynomial.

    Parameters:
    - X (list[PyFelt | ModuloCircuitElement]): Coefficients to evaluate.

    Returns:
    - list[int]: Sparsity categories for each coefficient in `X`.
    """
    return [2 if x.value == 1 else 1 if x.value != 0 else 0 for x in X]


def tower_to_direct(
    X: list[PyFelt], curve_id: int, extension_degree: int
) -> list[PyFelt]:
    """
    Convert Tower to Direct extension field representation:
    T(u,v) = t0+t1u + (t2+t3u)v + (t4+t5u)v^2
    D(v) = d0+d1v+d2v^2+d3v^3+d4v^4+d5v^5

    Only tested with BN254 and BLS12_381 6th and 12th towers defined in this file.
    They were computed by hand then abstracted away with no guarantee of genericity under different tower constructions.
    """
    assert len(X) == extension_degree and isinstance(
        X[0], PyFelt
    ), f"len(X)={len(X)}, type(X[0])={type(X[0])}"
    if extension_degree == 2:
        return X
    if extension_degree == 6:
        return TD6(X, curve_id)
    elif extension_degree == 12:
        return TD12(X, curve_id)
    else:
        raise ValueError(f"Unsupported extension degree {extension_degree}")


def direct_to_tower(
    X: list[PyFelt | ModuloCircuitElement], curve_id: int, extension_degree: int
) -> list[PyFelt]:
    """
    Convert Direct to Tower extension field representation:
    Input : D(v) = d0+d1v+d2v^2+d3v^3+d4v^4+d5v^5
    Output: T(u,v) = t0+t1u + (t2+t3u)v + (t4+t5u)v^2

    Only tested with BN254 and BLS12_381 6th and 12th towers defined in this file.
    They were computed by hand then abstracted away with no guarantee of genericity under different tower constructions.
    """
    assert len(X) == extension_degree and isinstance(
        X[0], (PyFelt, ModuloCircuitElement)
    ), f"{type(X[0])}, len(X)={len(X)}"
    X = [x.felt for x in X]
    if extension_degree == 2:
        return X
    if extension_degree == 6:
        return DT6(X, curve_id)
    elif extension_degree == 12:
        return DT12(X, curve_id)
    else:
        raise ValueError(f"Unsupported extension degree {extension_degree}")


def TD6(X: list[PyFelt], curve_id: int) -> list[PyFelt]:
    curve = CURVES[curve_id]
    assert isinstance(curve, PairingCurve)
    nr_a0 = curve.nr_a0
    return [
        X[0] - nr_a0 * X[1],
        X[2] - nr_a0 * X[3],
        X[4] - nr_a0 * X[5],
        X[1],
        X[3],
        X[5],
    ]


def DT6(X: list[PyFelt], curve_id: int) -> list[PyFelt]:
    curve = CURVES[curve_id]
    assert isinstance(curve, PairingCurve)
    nr_a0 = curve.nr_a0
    return [
        X[0] + nr_a0 * X[3],
        X[3],
        X[1] + nr_a0 * X[4],
        X[4],
        X[2] + nr_a0 * X[5],
        X[5],
    ]


def TD12(X: list[PyFelt], curve_id: int) -> list[PyFelt]:
    curve = CURVES[curve_id]
    assert isinstance(curve, PairingCurve)
    nr_a0 = curve.nr_a0
    return [
        X[0] - nr_a0 * X[1],
        X[6] - nr_a0 * X[7],
        X[2] - nr_a0 * X[3],
        X[8] - nr_a0 * X[9],
        X[4] - nr_a0 * X[5],
        X[10] - nr_a0 * X[11],
        X[1],
        X[7],
        X[3],
        X[9],
        X[5],
        X[11],
    ]


def DT12(X: list[PyFelt], curve_id: int) -> list[PyFelt]:
    X += (12 - len(X)) * [0]
    curve = CURVES[curve_id]
    assert isinstance(curve, PairingCurve)
    nr_a0 = curve.nr_a0
    return [
        X[0] + nr_a0 * X[6],
        X[6],
        X[2] + nr_a0 * X[8],
        X[8],
        X[4] + nr_a0 * X[10],
        X[10],
        X[1] + nr_a0 * X[7],
        X[7],
        X[3] + nr_a0 * X[9],
        X[9],
        X[5] + nr_a0 * X[11],
        X[11],
    ]


def precompute_lineline_sparsity(curve_id: int):
    curve = CURVES[curve_id]
    assert isinstance(curve, PairingCurve)
    field = get_base_field(curve_id)
    line_sparsity = curve.line_function_sparsity
    line = Polynomial([field(x) for x in line_sparsity])
    ll = line * line % get_irreducible_poly(curve_id, 12)
    ll_sparsity = get_sparsity(ll.coefficients)
    return ll_sparsity[0:12]


def replace_consecutive_zeros(lst):
    result = []
    i = 0
    while i < len(lst):
        if i < len(lst) - 1 and lst[i] == 0 and lst[i + 1] == 0:
            result.append(3)  # Replace consecutive zeros with 3
            i += 2
        elif lst[i] == -1:
            result.append(2)  # Replace -1 with 2
            i += 1
        else:
            result.append(lst[i])
            i += 1
    return result


def recode_naf_bits(lst):
    result = []
    i = 0
    while i < len(lst):
        if i < len(lst) - 1 and lst[i] == 0 and (lst[i + 1] == 1 or lst[i + 1] == -1):
            # "01" or "0-1"
            if lst[i + 1] == 1:
                result.append(3)  # Replace "01" with 3
            else:
                result.append(4)  # Replace "0-1" with 4
            i += 2
        elif i < len(lst) - 1 and (lst[i] == 1 or lst[i] == -1) and lst[i + 1] == 0:
            # "10" or "-10"
            if lst[i] == 1:
                result.append(1)  # Replace 10 with 6
            else:
                result.append(2)  # Replace -10 with 7
            i += 2
        elif i < len(lst) - 1 and lst[i] == 0 and lst[i + 1] == 0:
            result.append(0)  # Replace consecutive zeros with 0
            i += 2
        else:
            raise ValueError(f"Unexpected bit sequence at index {i}")
    return result


if __name__ == "__main__":
    r = recode_naf_bits(jy00(6 * 0x44E992B44A6909F1 + 2)[2:])
    print(r, len(r))
    from garaga.hints.io import int_to_u384

    # bls = [int(x) for x in bin(0xD201000000010000)[2:]][2:]
    # recode_naf_bits(bls)

    def print_nbits_and_nG_glv_fake_glv():
        for curve_id in CURVES:
            curve: WeierstrassCurve = CURVES[curve_id]
            if curve.is_endomorphism_available():
                nbits = curve.n.bit_length() // 4 + 9
                print(
                    f"Curve {curve_id}: {nbits}, {G1Point.get_nG(CurveID(curve_id), 2 ** (nbits - 1)).to_cairo_1()}"
                )

    print_nbits_and_nG_glv_fake_glv()

    def print_min_one_order():
        for curve_id in CURVES:
            curve: WeierstrassCurve = CURVES[curve_id]
            print(f"Curve {curve_id}: min_one_order: {int_to_u384(-1 % curve.n)}")

    print_min_one_order()

    print()

    def print_third_root_of_unity():
        for curve_id in CURVES:
            curve: WeierstrassCurve = CURVES[curve_id]
            if curve.is_endomorphism_available():
                print(
                    f"Curve {curve_id}: third_root_of_unity: {int_to_u384(curve.third_root_of_unity)}"
                )

    print_third_root_of_unity()



from dataclasses import dataclass
from functools import lru_cache

from garaga import garaga_rs
from garaga import modulo_circuit_structs as structs
from garaga.algebra import PyFelt
from garaga.definitions import (
    CURVES,
    STARK,
    CurveID,
    G1Point,
    WeierstrassCurve,
    get_base_field,
)
from garaga.hints import fake_glv


@dataclass(slots=True)
class MSMCalldataBuilder:
    curve_id: CurveID
    points: list[G1Point]
    scalars: list[int]

    def __post_init__(self):
        assert all(
            point.curve_id == self.curve_id for point in self.points
        ), "All points must be on the same curve."
        assert len(self.points) == len(
            self.scalars
        ), f"Number of points and scalars must be equal, got {len(self.points)} points and {len(self.scalars)} scalars"
        assert all(
            0 <= s <= CURVES[self.curve_id.value].n for s in self.scalars
        ), f"Scalars must be in [0, {self.curve_id.name}'s order] == [0, {CURVES[self.curve_id.value].n}]."

    def __hash__(self) -> int:
        return hash((self.curve_id, tuple(self.points), tuple(self.scalars)))

    @property
    def field(self):
        return get_base_field(self.curve_id)

    @property
    def msm_size(self):
        return len(self.scalars)

    @property
    def curve(self) -> WeierstrassCurve:
        return CURVES[self.curve_id.value]

    @lru_cache(maxsize=2)
    def build_msm_hint(self) -> structs.StructArray:
        """
        Returns the MSMHint
        """
        hints_structs = []
        for i, (point, scalar) in enumerate(zip(self.points, self.scalars)):
            if self.curve.is_endomorphism_available():
                Q, u1, u2, v1, v2 = fake_glv.get_glv_fake_glv_hint(point, scalar)
                hints_structs.append(
                    structs.Struct(
                        struct_name=f"GlvFakeGlvHint",
                        name=f"glv_fake_glv_hint_{i}",
                        elmts=[
                            structs.G1PointCircuit.from_G1Point("Q", Q),
                            structs.felt252(name=f"u1_{i}", elmts=[PyFelt(u1, STARK)]),
                            structs.felt252(name=f"u2_{i}", elmts=[PyFelt(u2, STARK)]),
                            structs.felt252(name=f"v1_{i}", elmts=[PyFelt(v1, STARK)]),
                            structs.felt252(name=f"v2_{i}", elmts=[PyFelt(v2, STARK)]),
                        ],
                    )
                )
            else:
                Q, s1, s2 = fake_glv.get_fake_glv_hint(point, scalar)
                hints_structs.append(
                    structs.Struct(
                        struct_name=f"FakeGlvHint",
                        name=f"fake_glv_hint_{i}",
                        elmts=[
                            structs.G1PointCircuit.from_G1Point("Q", Q),
                            structs.felt252(name=f"s1_{i}", elmts=[PyFelt(s1, STARK)]),
                            structs.felt252(name=f"s2_{i}", elmts=[PyFelt(s2, STARK)]),
                        ],
                    )
                )

        return structs.StructSpan(name="msm_hint", elmts=hints_structs)

    def _get_input_structs(
        self,
    ) -> list[structs.Cairo1SerializableStruct]:
        """ """
        inputs = []
        inputs.append(self.build_msm_hint())  # msm_hint
        inputs.append(
            structs.StructSpan(
                name="points",
                elmts=[
                    structs.G1PointCircuit.from_G1Point(f"point{i}", point)
                    for i, point in enumerate(self.points)
                ],
            )
        )
        inputs.append(
            structs.StructSpan(
                name="scalars",
                elmts=[
                    structs.u256(name=f"scalar{i}", elmts=[PyFelt(s, 2**256)])
                    for i, s in enumerate(self.scalars)
                ],
            )
        )

        return inputs

    def to_cairo_1_test(
        self, test_name: str = None, include_digits_decomposition=False
    ):
        # print(
        #     f"Generating MSM test for {self.curve_id.name} with {len(self.scalars)} points"
        # )
        test_name = test_name or f"test_msm_{self.curve_id.name}_{len(self.scalars)}P"
        inputs = garaga_rs.msm_calldata_builder(
            [value for point in self.points for value in [point.x, point.y]],
            self.scalars,
            self.curve_id.value,
            True,
            False,
        )
        Q = structs.G1PointCircuit.from_G1Point(
            "Q", G1Point.msm(self.points, self.scalars)
        )

        T = "u384" if self.curve_id == CurveID.BLS12_381 else "u288"
        code = f"""
        #[test]
        fn {test_name}() {{
            let mut data = array![{','.join([hex(value) for value in inputs])}].span();
            let points = Serde::deserialize(ref data).unwrap();
            let scalars = Serde::deserialize(ref data).unwrap();
            let curve_id = Serde::deserialize(ref data).unwrap();
            let res = msm_g1(points, scalars, curve_id, data);
            assert!(res == {Q.serialize(raw=True)});
        }}
        """
        return code

    def _serialize_to_calldata_rust(
        self,
        include_points_and_scalars=True,
        serialize_as_pure_felt252_array=False,
    ) -> list[int]:
        return garaga_rs.msm_calldata_builder(
            [value for point in self.points for value in [point.x, point.y]],
            self.scalars,
            self.curve_id.value,
            include_points_and_scalars,
            serialize_as_pure_felt252_array,
        )

    def serialize_to_calldata(
        self,
        include_points_and_scalars=True,
        serialize_as_pure_felt252_array=False,
        use_rust=False,
    ) -> list[int]:
        if use_rust:
            return self._serialize_to_calldata_rust(
                include_points_and_scalars,
                serialize_as_pure_felt252_array,
            )

        msm_hint, points, scalars = self._get_input_structs()

        call_data: list[int] = []

        if include_points_and_scalars:
            call_data.extend(points.serialize_to_calldata())
            call_data.extend(scalars.serialize_to_calldata())
            call_data.append(self.curve_id.value)

        call_data.extend(msm_hint.serialize_to_calldata()[1:])

        if serialize_as_pure_felt252_array:
            return [len(call_data)] + call_data
        else:
            return call_data


if __name__ == "__main__":
    import random

    c = CurveID.SECP256K1
    order = CURVES[c.value].n
    msm = MSMCalldataBuilder(
        curve_id=c,
        points=[G1Point.gen_random_point(c) for _ in range(1)],
        scalars=[random.randint(0, order) for _ in range(1)],
    )
    cd = msm.serialize_to_calldata(
        include_points_and_scalars=True,
        serialize_as_pure_felt252_array=False,
    )
    print(cd)
    print(len(cd))



from typing import Optional

import typer

from garaga.starknet.cli.declare import declare
from garaga.starknet.cli.deploy import deploy
from garaga.starknet.cli.gen import gen
from garaga.starknet.cli.utils import get_package_version
from garaga.starknet.cli.verify import calldata, verify_onchain


def version_callback(value: bool):
    if value:
        print(f"garaga version: {get_package_version()}")
        raise typer.Exit()


app = typer.Typer(
    no_args_is_help=True,  # Show help when no arguments are provided
    context_settings={"help_option_names": ["-h", "--help"]},
)


# Add version option that will be processed first
@app.callback(invoke_without_command=True)
def main(
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        callback=version_callback,
        is_eager=True,
        help="Show the version and exit.",
    ),
):
    pass


app.command(no_args_is_help=True)(gen)
app.command(no_args_is_help=True)(declare)
app.command(no_args_is_help=True)(deploy)
app.command(no_args_is_help=True)(verify_onchain)
app.command(no_args_is_help=True)(calldata)


if __name__ == "__main__":
    app()



import asyncio
from pathlib import Path
from typing import Annotated, Optional

import rich
import typer
from dotenv import load_dotenv
from starknet_py.contract import Contract
from starknet_py.hash.address import compute_address
from starknet_py.hash.utils import pedersen_hash

from garaga.hints.io import to_hex_str, to_int
from garaga.starknet.cli.utils import (
    Network,
    get_contract_if_exists,
    get_contract_if_exists_async,
    load_account,
)

# Mainnet/Sepolia deployer address
DEPLOYER_ADDRESS = 0x41A78E741E5AF2FEC34B695679BC6891742439F7AFB8484ECD7766661AD02BF

app = typer.Typer()


async def perform_contract_deployment(
    account, class_hash: str, salt: int, precomputed_address: int
) -> Optional[Contract]:
    """Attempt to deploy a contract with the given parameters."""
    try:
        deploy_result = await Contract.deploy_contract_v3(
            account=account,
            class_hash=class_hash,
            deployer_address=DEPLOYER_ADDRESS,
            auto_estimate=True,
            salt=salt,
            cairo_version=1,
        )
        deploy_result = await deploy_result.wait_for_acceptance()
        return deploy_result.deployed_contract
    except Exception as e:
        print("Waiting for transaction to be accepted...")
        await asyncio.sleep(5.5)
        contract = await get_contract_if_exists_async(account, precomputed_address)
        if contract:
            return contract
        else:
            rich.print(f"[bold red]Deployment error: {e}[/bold red]")
        return None


def print_contract_info(contract: Contract, network: Network):
    """Print information about the deployed contract."""
    sepolia_prefix = "" if network == Network.MAINNET else "sepolia."
    rich.print(f"Contract address on {network.name}: {hex(contract.address)}")
    rich.print(
        f"Voyager link: https://{sepolia_prefix}voyager.online/contract/{hex(contract.address)}"
    )
    rich.print(f"Available endpoints: {list(contract.functions.keys())}")


def deploy(
    class_hash: Annotated[
        str,
        typer.Option(
            help="Contract class hash to deploy. Can be decimal or hex string"
        ),
    ],
    env_file: Annotated[
        Path,
        typer.Option(
            help="Path to the environment file containing rpc, address, private_key",
            file_okay=True,
            dir_okay=False,
            exists=True,
            autocompletion=lambda: [],
        ),
    ] = str(Path.cwd() / ".secrets"),
    network: Annotated[
        Network,
        typer.Option(
            help="Starknet network",
            case_sensitive=False,
        ),
    ] = Network.SEPOLIA.value,
    salt: Annotated[
        int,
        typer.Option(
            help="Salt value for contract address calculation (default: 1)",
        ),
    ] = 1,
):
    """Deploy an instance of a smart contract class hash to Starknet. Obtain its address, the available endpoints and a explorer link."""

    # Validate and load environment
    if not Path(env_file).exists():
        raise FileNotFoundError(f"Environment file {env_file} does not exist")
    load_dotenv(env_file)
    # Prepare account and compute address
    account = load_account(network)
    class_hash = to_hex_str(class_hash)

    precomputed_address = compute_address(
        class_hash=to_int(class_hash),
        constructor_calldata=[],
        salt=pedersen_hash(to_int(account.address), salt),
        deployer_address=DEPLOYER_ADDRESS,
    )

    # Check if contract already exists
    contract = get_contract_if_exists(account, precomputed_address)
    if contract:
        rich.print(
            f"[bold green]Contract already deployed at {hex(contract.address)}[/bold green]"
        )
        print_contract_info(contract, network)
        return

    # Perform deployment
    rich.print(f"Deploying contract with class hash {class_hash}...")
    contract = asyncio.run(
        perform_contract_deployment(account, class_hash, salt, precomputed_address)
    )

    # Handle deployment result
    if not contract:
        # Check if the contract exists despite the error (transaction might have completed)
        contract = get_contract_if_exists(account, precomputed_address)
        if contract:
            rich.print(
                f"[bold green]Contract found at {hex(contract.address)} despite deployment errors.[/bold green]"
            )
        else:
            rich.print(
                f"[bold yellow]Contract not deployed. Try running the command again - it might be a temporary network issue.[/bold yellow]"
            )
            return
    else:
        rich.print(
            f"[bold green]Contract successfully deployed at {hex(contract.address)}[/bold green]"
        )

    # Print contract information
    print_contract_info(contract, network)



import concurrent.futures
import os
import subprocess
from pathlib import Path

from garaga.definitions import CurveID
from garaga.modulo_circuit_structs import G2Line, StructArray
from garaga.precompiled_circuits.multi_miller_loop import precompute_lines
from garaga.starknet.cli.utils import create_directory, get_package_version
from garaga.starknet.groth16_contract_generator.parsing_utils import Groth16VerifyingKey

ECIP_OPS_CLASS_HASH = 0x465991EC820CF53DBB2B27474B6663FB6F0C8BF3DAC7DB3991960214FAD97F5
CAIRO_VERSION = "2.11.4"


def precompute_lines_from_vk(vk: Groth16VerifyingKey) -> StructArray:

    # Precompute lines for fixed G2 points
    lines = precompute_lines([vk.gamma, vk.delta])
    precomputed_lines = StructArray(
        name="lines",
        elmts=[
            G2Line(name=f"line{i}", elmts=lines[i : i + 4])
            for i in range(0, len(lines), 4)
        ],
    )

    return precomputed_lines


def gen_groth16_verifier(
    vk: str | Path | Groth16VerifyingKey,
    output_folder_path: str,
    output_folder_name: str,
    ecip_class_hash: int = ECIP_OPS_CLASS_HASH,
    cli_mode: bool = False,
) -> str:
    if isinstance(vk, (Path, str)):
        vk = Groth16VerifyingKey.from_json(vk)
    else:
        vk = vk

    curve_id = vk.curve_id
    if cli_mode:
        output_folder_name = output_folder_name
    else:
        output_folder_name = output_folder_name + f"_{curve_id.name.lower()}"
    output_folder_path = os.path.join(output_folder_path, output_folder_name)

    precomputed_lines = precompute_lines_from_vk(vk)

    constants_code = f"""
    use garaga::definitions::{{G1Point, G2Point, E12D, G2Line, u384}};
    {f"use garaga::definitions::u288;" if curve_id!=CurveID.BLS12_381 else ""}
    use garaga::groth16::Groth16VerifyingKey;

    pub const N_PUBLIC_INPUTS:usize = {len(vk.ic)-1};
    {vk.serialize_to_cairo()}
    pub const precomputed_lines: [G2Line; {len(precomputed_lines)//4}] = {precomputed_lines.serialize(raw=True, const=True)};
    """
    contract_code = f"""
use super::groth16_verifier_constants::{{N_PUBLIC_INPUTS, vk, ic, precomputed_lines}};

#[starknet::interface]
trait IGroth16Verifier{curve_id.name}<TContractState> {{
    fn verify_groth16_proof_{curve_id.name.lower()}(
        self: @TContractState,
        full_proof_with_hints: Span<felt252>,
    ) -> Option<Span<u256>>;
}}

#[starknet::contract]
mod Groth16Verifier{curve_id.name} {{
    use starknet::SyscallResultTrait;
    use garaga::definitions::{{G1Point, G1G2Pair}};
    use garaga::groth16::{{multi_pairing_check_{curve_id.name.lower()}_3P_2F_with_extra_miller_loop_result}};
    use garaga::ec_ops::{{G1PointTrait, ec_safe_add}};
    use garaga::ec_ops_g2::{{G2PointTrait}};
    use garaga::utils::calldata::{{deserialize_full_proof_with_hints_{curve_id.name.lower()}}};
    use super::{{N_PUBLIC_INPUTS, vk, ic, precomputed_lines}};

    const ECIP_OPS_CLASS_HASH: felt252 = {hex(ecip_class_hash)};

    #[storage]
    struct Storage {{}}

    #[abi(embed_v0)]
    impl IGroth16Verifier{curve_id.name} of super::IGroth16Verifier{curve_id.name}<ContractState> {{
        fn verify_groth16_proof_{curve_id.name.lower()}(
            self: @ContractState,
            full_proof_with_hints: Span<felt252>,
        ) -> Option<Span<u256>> {{
            // DO NOT EDIT THIS FUNCTION UNLESS YOU KNOW WHAT YOU ARE DOING.
            // This function returns an Option for the public inputs if the proof is valid.
            // If the proof is invalid, the execution will either fail or return None.
            // Read the documentation to learn how to generate the full_proof_with_hints array given a proof and a verifying key.
            let fph = deserialize_full_proof_with_hints_{curve_id.name.lower()}(full_proof_with_hints);
            let groth16_proof = fph.groth16_proof;
            let mpcheck_hint = fph.mpcheck_hint;
            let small_Q = fph.small_Q;
            let msm_hint = fph.msm_hint;

            groth16_proof.a.assert_on_curve({curve_id.value});
            groth16_proof.b.assert_on_curve({curve_id.value});
            groth16_proof.c.assert_on_curve({curve_id.value});

            let ic = ic.span();

            let vk_x: G1Point = match ic.len() {{
                0 => panic!("Malformed VK"),
                1 => *ic.at(0),
                _ => {{
                    // Start serialization with the hint array directly to avoid copying it.
                    let mut msm_calldata: Array<felt252> = array![];
                    // Add the points from VK and public inputs to the proof.
                    Serde::serialize(@ic.slice(1, N_PUBLIC_INPUTS), ref msm_calldata);
                    Serde::serialize(@groth16_proof.public_inputs, ref msm_calldata);
                    // Complete with the curve indentifier ({curve_id.value} for {curve_id.name}):
                    msm_calldata.append({curve_id.value});
                    // Add the hint array.
                    for x in msm_hint {{
                        msm_calldata.append(*x);
                    }}

                    // Call the multi scalar multiplication endpoint on the Garaga ECIP ops contract
                    // to obtain vk_x.
                    let mut _vx_x_serialized = starknet::syscalls::library_call_syscall(
                        ECIP_OPS_CLASS_HASH.try_into().unwrap(),
                        selector!("msm_g1"),
                        msm_calldata.span()
                    )
                        .unwrap_syscall();

                    ec_safe_add(
                        Serde::<G1Point>::deserialize(ref _vx_x_serialized).unwrap(), *ic.at(0), {curve_id.value}
                    )
                }}
            }};
            // Perform the pairing check.
            let check = multi_pairing_check_{curve_id.name.lower()}_3P_2F_with_extra_miller_loop_result(
                G1G2Pair {{ p: vk_x, q: vk.gamma_g2 }},
                G1G2Pair {{ p: groth16_proof.c, q: vk.delta_g2 }},
                G1G2Pair {{ p: groth16_proof.a.negate({curve_id.value}), q: groth16_proof.b }},
                vk.alpha_beta_miller_loop_result,
                precomputed_lines.span(),
                mpcheck_hint,
                small_Q
            );
            if check == true {{
                return Option::Some(groth16_proof.public_inputs);
            }} else {{
                return Option::None;
            }}
        }}
    }}
}}


    """

    create_directory(output_folder_path)
    src_dir = os.path.join(output_folder_path, "src")
    create_directory(src_dir)

    with open(os.path.join(output_folder_path, ".tool-versions"), "w") as f:
        f.write(f"scarb {CAIRO_VERSION}\n")

    with open(os.path.join(src_dir, "groth16_verifier_constants.cairo"), "w") as f:
        f.write(constants_code)

    with open(os.path.join(src_dir, "groth16_verifier.cairo"), "w") as f:
        f.write(contract_code)

    with open(os.path.join(output_folder_path, "Scarb.toml"), "w") as f:
        f.write(get_scarb_toml_file(output_folder_name, cli_mode))

    with open(os.path.join(src_dir, "lib.cairo"), "w") as f:
        f.write(
            """
mod groth16_verifier;
mod groth16_verifier_constants;
"""
        )
    subprocess.run(["scarb", "fmt", f"{output_folder_path}"], check=True)
    return constants_code


def get_scarb_toml_file(package_name: str, cli_mode: bool, inlining_level: int = 2):
    version = get_package_version()
    if version == "dev":
        suffix = ""
    else:
        suffix = ', tag = "v' + version + '"'
    if cli_mode:
        dep = 'git = "https://github.com/keep-starknet-strange/garaga.git"' + suffix
    else:
        dep = 'path = "../../../"'

    return f"""[package]
name = "{package_name}"
version = "0.1.0"
edition = "2024_07"

[dependencies]
garaga = {{ {dep} }}
starknet = "{CAIRO_VERSION}"

[cairo]
sierra-replace-ids = false
inlining-strategy = {inlining_level}

[dev-dependencies]
cairo_test = "{CAIRO_VERSION}"

[[target.starknet-contract]]
casm = true
casm-add-pythonic-hints = true
"""


if __name__ == "__main__":
    vk_paths = [
        "hydra/garaga/starknet/groth16_contract_generator/examples/vk_bn254.json",
        "hydra/garaga/starknet/groth16_contract_generator/examples/vk_bls.json",
    ]

    CONTRACTS_FOLDER = "src/contracts/autogenerated/"  # Do not change this

    FOLDER_NAME = "groth16_example"  # '_curve_id' is appended in the end.

    def _generate_verifier(vk_path):
        try:
            gen_groth16_verifier(
                vk_path, CONTRACTS_FOLDER, FOLDER_NAME, ECIP_OPS_CLASS_HASH
            )
        except Exception as e:
            print(f"An error occurred: {e}")

    with concurrent.futures.ProcessPoolExecutor(max_workers=2) as executor:
        executor.map(_generate_verifier, vk_paths)



from abc import abstractmethod
from typing import Dict, Tuple, Type, Union

import garaga.modulo_circuit_structs as structs
import garaga.precompiled_circuits.honk as hk
from garaga.definitions import CurveID
from garaga.modulo_circuit import ModuloCircuitElement
from garaga.modulo_circuit_structs import u384
from garaga.precompiled_circuits.compilable_circuits.base import (
    BaseModuloCircuit,
    ModuloCircuit,
    PyFelt,
)
from garaga.precompiled_circuits.honk import (
    PAIRING_POINT_OBJECT_LENGTH,
    ZK_BATCHED_RELATION_PARTIAL_LENGTH,
    HonkVerifierCircuits,
    HonkVk,
    Wire,
    ZKHonkVerifierCircuits,
)
from garaga.starknet.honk_contract_generator.calldata import filter_msm_scalars


class BaseUltraHonkCircuit(BaseModuloCircuit):
    def __init__(
        self,
        name: str,
        log_n: int,
        curve_id: CurveID.GRUMPKIN.value,
        auto_run: bool = True,
        compilation_mode: int = 1,
    ) -> None:
        self.log_n = log_n
        super().__init__(
            name=name,
            curve_id=curve_id,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    def _initialize_circuit(self) -> HonkVerifierCircuits:
        return HonkVerifierCircuits(
            name=self.name,
            log_n=self.log_n,
            curve_id=self.curve_id,
        )

    def _process_input(
        self, circuit: HonkVerifierCircuits, input: list[PyFelt]
    ) -> dict:
        """
        Method responsible for deserializing the input list of elements into the variables in the input map,
        and writing them to the circuit.
        The input list is expected to be in the same order as the input map.
        Since we use Python 3.10, the input map dict is ordered.
        Returns a vars dict with the same keys as the input map, but with the values being the instances of the structs,
        each struct holding ModuloCircuitElement(s).
        """
        vars = {}
        for name, struct_info in self.input_map.items():
            if name == "sumcheck_evaluations":
                # Edge case, TABLE_SHIFTS are not used. Replace with mocked non part of circuit values.
                struct_type, size = struct_info
                assert size == len(Wire) - len(Wire.unused_indexes())
                elements = circuit.write_struct(
                    struct_type(name, elmts=[input.pop(0) for _ in range(size)])
                )
                elements = Wire.insert_unused_indexes_with_nones(elements)
                assert len(elements) == len(Wire)
                vars[name] = elements

            elif isinstance(struct_info, tuple) and struct_info[0] in [
                structs.u256Span,
                structs.u384Array,
                structs.u128Span,
            ]:
                # Handle u384Array with specified size
                struct_type, size = struct_info
                vars[name] = circuit.write_struct(
                    struct_type(name, elmts=[input.pop(0) for _ in range(size)])
                )

            else:
                struct_type = struct_info
                # For other types, create a temporary instance to get its length
                temp_instance = struct_type(
                    name, None
                )  # Temporary instance with a dummy element
                size = len(temp_instance)
                if size == 1:
                    vars[name] = circuit.write_struct(
                        struct_type(name, elmts=[input.pop(0)])
                    )
                else:
                    # Convert to tuple so that it is Hashable
                    vars[name] = tuple(
                        circuit.write_struct(
                            struct_type(name, elmts=[input.pop(0) for _ in range(size)])
                        )
                    )
        assert len(input) == 0, f"Expected input of length 0, got {len(input)}"

        return vars

    @property
    @abstractmethod
    def input_map(
        self,
    ) -> Dict[
        str,
        Union[
            Type[structs.Cairo1SerializableStruct],
            Tuple[Type[structs.Cairo1SerializableStruct], int],
        ],
    ]:
        """
        Define the input map for the circuit in a dict.
        The key will be the name of the input variable, also used in the signature of the compiled Cairo code.
        The value will be either a Cairo1SerializableStruct type (which defines the struct in the Cairo code),
            or a tuple of the type and its size (for Array-like types).
        The reason behind this is that each Cairo1SerializableStruct defines the __len__ method, but for the
        array-like structs we need to specify the size in advance.
        """

    def build_input(self) -> list[PyFelt]:
        """
        Extends the base method of BaseModuloCircuit, by reading the input map and returning a list of random elements of the total expected size.
        """
        total_elements = 0
        for name, struct_info in self.input_map.items():
            if isinstance(struct_info, tuple):
                # Array-like case
                _, size = struct_info
                total_elements += size
            else:
                # For other structs, create a temporary instance to get its length
                temp_instance = struct_info(name, None)
                total_elements += len(temp_instance)

        # print(f"Total elements: {total_elements}")
        return [self.field.random() for _ in range(total_elements)]

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = self._initialize_circuit()

        vars = self._process_input(circuit, input)
        return self._execute_circuit_logic(circuit, vars)

    @abstractmethod
    def _execute_circuit_logic(self, circuit, vars) -> ModuloCircuit:
        """
        Implement the circuit logic using the processed input variables.
        """


class SumCheckCircuit(BaseUltraHonkCircuit):
    def __init__(
        self,
        vk: HonkVk,
        curve_id: int = CurveID.GRUMPKIN.value,
        auto_run: bool = True,
        compilation_mode: int = 1,
    ) -> None:
        assert vk.public_inputs_size >= PAIRING_POINT_OBJECT_LENGTH
        name = f"honk_sumcheck_size_{vk.log_circuit_size}_pub_{vk.public_inputs_size}"
        self.vk = vk

        super().__init__(
            name, vk.log_circuit_size, curve_id, auto_run, compilation_mode
        )

    @property
    def input_map(self) -> dict:
        imap = {}

        imap["p_public_inputs"] = (
            structs.u256Span,
            self.vk.public_inputs_size - PAIRING_POINT_OBJECT_LENGTH,
        )
        imap["p_pairing_point_object"] = (structs.u256Span, PAIRING_POINT_OBJECT_LENGTH)
        imap["p_public_inputs_offset"] = structs.u384

        imap["sumcheck_univariates_flat"] = (
            structs.u256Span,
            self.vk.log_circuit_size * hk.BATCHED_RELATION_PARTIAL_LENGTH,
        )

        imap["sumcheck_evaluations"] = (
            structs.u256Span,
            hk.NUMBER_OF_ENTITIES - len(Wire.unused_indexes()),
        )

        imap["tp_sum_check_u_challenges"] = (
            structs.u128Span,
            self.vk.log_circuit_size,
        )
        imap["tp_gate_challenges"] = (
            structs.u128Span,
            self.vk.log_circuit_size,
        )
        imap["tp_eta_1"] = structs.u128
        imap["tp_eta_2"] = structs.u128
        imap["tp_eta_3"] = structs.u128
        imap["tp_beta"] = structs.u128
        imap["tp_gamma"] = structs.u128
        imap["tp_base_rlc"] = structs.u384
        imap["tp_alphas"] = (structs.u128Span, hk.NUMBER_OF_ALPHAS)
        return imap

    def _execute_circuit_logic(
        self, circuit: HonkVerifierCircuits, vars: dict
    ) -> ModuloCircuit:

        tp_delta = circuit.compute_public_input_delta(
            vars["p_public_inputs"],
            vars["p_pairing_point_object"],
            vars["tp_beta"],
            vars["tp_gamma"],
            self.vk.circuit_size,
            vars["p_public_inputs_offset"],
        )

        sumcheck_univariates_flat = vars["sumcheck_univariates_flat"]
        sumcheck_univariates = []
        for i in range(self.vk.log_circuit_size):
            sumcheck_univariates.append(
                sumcheck_univariates_flat[
                    i
                    * hk.BATCHED_RELATION_PARTIAL_LENGTH : (i + 1)
                    * hk.BATCHED_RELATION_PARTIAL_LENGTH
                ]
            )

        assert len(sumcheck_univariates) == self.vk.log_circuit_size
        assert len(sumcheck_univariates[0]) == hk.BATCHED_RELATION_PARTIAL_LENGTH

        assert len(vars["sumcheck_evaluations"]) == len(Wire)

        check_rlc, check = circuit.verify_sum_check(
            sumcheck_univariates,
            vars["sumcheck_evaluations"],
            vars["tp_beta"],
            vars["tp_gamma"],
            tp_delta,
            vars["tp_eta_1"],
            vars["tp_eta_2"],
            vars["tp_eta_3"],
            vars["tp_sum_check_u_challenges"],
            vars["tp_gate_challenges"],
            vars["tp_alphas"],
            self.vk.log_circuit_size,
            vars["tp_base_rlc"],
        )

        assert type(check_rlc) == ModuloCircuitElement
        assert type(check) == ModuloCircuitElement

        circuit.extend_struct_output(u384("check_rlc", elmts=[check_rlc]))
        circuit.extend_struct_output(u384("check", elmts=[check]))
        return circuit


class PrepareScalarsCircuit(BaseUltraHonkCircuit):
    def __init__(
        self,
        vk: HonkVk,
        curve_id: int = CurveID.GRUMPKIN.value,
        auto_run: bool = True,
        compilation_mode: int = 1,
    ) -> None:
        name = f"honk_prep_msm_scalars_size_{vk.log_circuit_size}"
        self.vk = vk
        self.scalar_indexes = []
        super().__init__(
            name, vk.log_circuit_size, curve_id, auto_run, compilation_mode
        )

    @property
    def input_map(self) -> dict:
        imap = {}

        imap["p_sumcheck_evaluations"] = (structs.u256Span, hk.NUMBER_OF_ENTITIES)
        imap["p_gemini_a_evaluations"] = (structs.u256Span, self.vk.log_circuit_size)
        imap["tp_gemini_r"] = structs.u384
        imap["tp_rho"] = structs.u384
        imap["tp_shplonk_z"] = structs.u384
        imap["tp_shplonk_nu"] = structs.u384
        imap["tp_sum_check_u_challenges"] = (
            structs.u128Span,
            self.vk.log_circuit_size,
        )

        return imap

    def _execute_circuit_logic(
        self, circuit: HonkVerifierCircuits, vars: dict
    ) -> ModuloCircuit:

        assert len(vars["p_sumcheck_evaluations"]) == len(Wire)

        scalars = circuit.compute_shplemini_msm_scalars(
            vars["p_sumcheck_evaluations"],
            vars["p_gemini_a_evaluations"],
            vars["tp_gemini_r"],
            vars["tp_rho"],
            vars["tp_shplonk_z"],
            vars["tp_shplonk_nu"],
            vars["tp_sum_check_u_challenges"],
        )

        scalars_filtered_no_nones = filter_msm_scalars(
            scalars, self.vk.log_circuit_size, False
        )
        # Remove shplonk_z (last scalar) (included in transcript)
        scalars_filtered_no_nones = scalars_filtered_no_nones[:-1]

        # sum_scalars = circuit.sum(scalars_filtered_no_nones)

        # For each filtered scalar, find its original index by matching offset
        self.scalar_indexes = []
        for scalar in scalars_filtered_no_nones:
            original_index = next(
                i
                for i, orig_scalar in enumerate(scalars)
                if orig_scalar is not None and orig_scalar.offset == scalar.offset
            )
            self.scalar_indexes.append(original_index)
            circuit.extend_struct_output(
                u384(f"scalar_{original_index}", elmts=[scalar])
            )

        self.msm_len = len(scalars_filtered_no_nones) + 1

        # circuit.extend_struct_output(u384("sum_scalars", elmts=[sum_scalars]))
        # circuit.exact_output_refs_needed = [sum_scalars]

        return circuit


class ZKBaseUltraHonkCircuit(BaseUltraHonkCircuit):
    def __init__(
        self,
        name: str,
        log_n: int,
        curve_id: CurveID.GRUMPKIN.value,
        auto_run: bool = True,
        compilation_mode: int = 1,
    ) -> None:
        super().__init__(
            name=name,
            log_n=log_n,
            curve_id=curve_id,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    def _initialize_circuit(self) -> HonkVerifierCircuits:
        return ZKHonkVerifierCircuits(
            name=self.name,
            log_n=self.log_n,
            curve_id=self.curve_id,
        )


class ZKSumCheckCircuit(ZKBaseUltraHonkCircuit):
    def __init__(
        self,
        vk: HonkVk,
        curve_id: int = CurveID.GRUMPKIN.value,
        auto_run: bool = True,
        compilation_mode: int = 1,
    ) -> None:
        assert (
            vk.public_inputs_size >= PAIRING_POINT_OBJECT_LENGTH
        ), f"vk.public_inputs_size: {vk.public_inputs_size}, PAIRING_POINT_OBJECT_LENGTH: {PAIRING_POINT_OBJECT_LENGTH}"
        name = (
            f"zk_honk_sumcheck_size_{vk.log_circuit_size}_pub_{vk.public_inputs_size}"
        )
        self.vk = vk

        super().__init__(
            name, vk.log_circuit_size, curve_id, auto_run, compilation_mode
        )

    @property
    def input_map(self) -> dict:
        imap = {}

        imap["p_public_inputs"] = (
            structs.u256Span,
            self.vk.public_inputs_size - PAIRING_POINT_OBJECT_LENGTH,
        )
        imap["p_pairing_point_object"] = (structs.u256Span, PAIRING_POINT_OBJECT_LENGTH)
        imap["p_public_inputs_offset"] = structs.u384

        imap["libra_sum"] = structs.u384

        imap["sumcheck_univariates_flat"] = (
            structs.u256Span,
            self.vk.log_circuit_size * ZK_BATCHED_RELATION_PARTIAL_LENGTH,
        )

        imap["sumcheck_evaluations"] = (
            structs.u256Span,
            hk.NUMBER_OF_ENTITIES - len(Wire.unused_indexes()),
        )

        imap["libra_evaluation"] = structs.u384

        imap["tp_sum_check_u_challenges"] = (
            structs.u128Span,
            self.vk.log_circuit_size,
        )
        imap["tp_gate_challenges"] = (
            structs.u128Span,
            self.vk.log_circuit_size,
        )
        imap["tp_eta_1"] = structs.u384
        imap["tp_eta_2"] = structs.u384
        imap["tp_eta_3"] = structs.u384
        imap["tp_beta"] = structs.u384
        imap["tp_gamma"] = structs.u384
        imap["tp_base_rlc"] = structs.u384
        imap["tp_alphas"] = (structs.u128Span, hk.NUMBER_OF_ALPHAS)
        imap["tp_libra_challenge"] = structs.u384
        return imap

    def _execute_circuit_logic(
        self, circuit: ZKHonkVerifierCircuits, vars: dict
    ) -> ModuloCircuit:

        tp_delta = circuit.compute_public_input_delta(
            vars["p_public_inputs"],
            vars["p_pairing_point_object"],
            vars["tp_beta"],
            vars["tp_gamma"],
            self.vk.circuit_size,
            vars["p_public_inputs_offset"],
        )

        sumcheck_univariates_flat = vars["sumcheck_univariates_flat"]
        sumcheck_univariates = []
        for i in range(self.vk.log_circuit_size):
            sumcheck_univariates.append(
                sumcheck_univariates_flat[
                    i
                    * ZK_BATCHED_RELATION_PARTIAL_LENGTH : (i + 1)
                    * ZK_BATCHED_RELATION_PARTIAL_LENGTH
                ]
            )

        assert len(sumcheck_univariates) == self.vk.log_circuit_size
        assert len(sumcheck_univariates[0]) == ZK_BATCHED_RELATION_PARTIAL_LENGTH

        assert len(vars["sumcheck_evaluations"]) == len(Wire)

        check_rlc, check = circuit.verify_sum_check(
            vars["libra_sum"],
            sumcheck_univariates,
            vars["sumcheck_evaluations"],
            vars["libra_evaluation"],
            vars["tp_beta"],
            vars["tp_gamma"],
            tp_delta,
            vars["tp_eta_1"],
            vars["tp_eta_2"],
            vars["tp_eta_3"],
            vars["tp_libra_challenge"],
            vars["tp_sum_check_u_challenges"],
            vars["tp_gate_challenges"],
            vars["tp_alphas"],
            self.vk.log_circuit_size,
            vars["tp_base_rlc"],
        )

        assert type(check_rlc) == ModuloCircuitElement
        assert type(check) == ModuloCircuitElement

        circuit.extend_struct_output(u384("check_rlc", elmts=[check_rlc]))
        circuit.extend_struct_output(u384("check", elmts=[check]))
        return circuit


class ZKPrepareScalarsCircuit(ZKBaseUltraHonkCircuit):
    def __init__(
        self,
        vk: HonkVk,
        curve_id: int = CurveID.GRUMPKIN.value,
        auto_run: bool = True,
        compilation_mode: int = 1,
    ) -> None:
        name = f"zkhonk_prep_msm_scalars_size_{vk.log_circuit_size}"
        self.vk = vk
        self.scalar_indexes = []
        super().__init__(
            name, vk.log_circuit_size, curve_id, auto_run, compilation_mode
        )

    @property
    def input_map(self) -> dict:
        imap = {}

        imap["p_sumcheck_evaluations"] = (structs.u256Span, hk.NUMBER_OF_ENTITIES)
        imap["p_gemini_masking_eval"] = structs.u384
        imap["p_gemini_a_evaluations"] = (structs.u256Span, self.vk.log_circuit_size)
        imap["p_libra_poly_evals"] = (structs.u256Span, 4)
        imap["tp_gemini_r"] = structs.u384
        imap["tp_rho"] = structs.u384
        imap["tp_shplonk_z"] = structs.u384
        imap["tp_shplonk_nu"] = structs.u384
        imap["tp_sum_check_u_challenges"] = (
            structs.u128Span,
            self.vk.log_circuit_size,
        )

        return imap

    def _execute_circuit_logic(
        self, circuit: ZKHonkVerifierCircuits, vars: dict
    ) -> ModuloCircuit:

        assert len(vars["p_sumcheck_evaluations"]) == len(Wire)

        scalars = circuit.compute_shplemini_msm_scalars(
            vars["p_sumcheck_evaluations"],
            vars["p_gemini_masking_eval"],
            vars["p_gemini_a_evaluations"],
            vars["p_libra_poly_evals"],
            vars["tp_gemini_r"],
            vars["tp_rho"],
            vars["tp_shplonk_z"],
            vars["tp_shplonk_nu"],
            vars["tp_sum_check_u_challenges"],
        )

        scalars_filtered_no_nones = filter_msm_scalars(
            scalars, self.vk.log_circuit_size, True
        )
        # Remove shplonk_z (last scalar) (included in transcript)
        scalars_filtered_no_nones = scalars_filtered_no_nones[:-1]
        self.msm_len = len(scalars_filtered_no_nones) + 1

        # For each filtered scalar, find its original index by matching offset
        self.scalar_indexes = []
        for scalar in scalars_filtered_no_nones:
            original_index = next(
                i
                for i, orig_scalar in enumerate(scalars)
                if orig_scalar is not None and orig_scalar.offset == scalar.offset
            )
            self.scalar_indexes.append(original_index)
            circuit.extend_struct_output(
                u384(f"scalar_{original_index}", elmts=[scalar])
            )

        self.msm_len = len(scalars_filtered_no_nones) + 1

        return circuit


class ZKEvalsConsistencyCircuit(ZKBaseUltraHonkCircuit):
    def __init__(
        self,
        vk: HonkVk,
        curve_id: int = CurveID.GRUMPKIN.value,
        auto_run: bool = True,
        compilation_mode: int = 1,
    ) -> None:
        name = f"zk_honk_evals_consist_size_{vk.log_circuit_size}"
        self.vk = vk
        super().__init__(
            name, vk.log_circuit_size, curve_id, auto_run, compilation_mode
        )

    @property
    def input_map(self) -> dict:
        imap = {}

        imap["p_libra_evaluation"] = structs.u384
        imap["p_libra_poly_evals"] = (structs.u256Span, 4)
        imap["tp_gemini_r"] = structs.u384
        imap["tp_sum_check_u_challenges"] = (
            structs.u128Span,
            hk.CONST_PROOF_SIZE_LOG_N,
        )

        return imap

    def _execute_circuit_logic(
        self, circuit: ZKHonkVerifierCircuits, vars: dict
    ) -> ModuloCircuit:

        vanishing_check, diff_check = circuit.check_evals_consistency(
            vars["p_libra_evaluation"],
            vars["p_libra_poly_evals"],
            vars["tp_gemini_r"],
            vars["tp_sum_check_u_challenges"],
        )

        assert type(vanishing_check) == ModuloCircuitElement
        assert type(diff_check) == ModuloCircuitElement

        circuit.extend_struct_output(u384("vanishing_check", elmts=[vanishing_check]))
        circuit.extend_struct_output(u384("diff_check", elmts=[diff_check]))
        return circuit


class ZKEvalsConsistencyInitCircuit(ZKBaseUltraHonkCircuit):
    def __init__(
        self,
        vk: HonkVk,
        curve_id: int = CurveID.GRUMPKIN.value,
        auto_run: bool = True,
        compilation_mode: int = 1,
    ) -> None:
        name = f"zk_honk_evals_cons_init_size_{vk.log_circuit_size}"
        self.vk = vk
        super().__init__(
            name, vk.log_circuit_size, curve_id, auto_run, compilation_mode
        )

    @property
    def input_map(self) -> dict:
        imap = {}

        imap["tp_gemini_r"] = structs.u384

        return imap

    def _execute_circuit_logic(
        self, circuit: ZKHonkVerifierCircuits, vars: dict
    ) -> ModuloCircuit:

        challenge_poly_eval, root_power_times_tp_gemini_r = (
            circuit._check_evals_consistency_init(
                vars["tp_gemini_r"],
            )
        )

        assert type(challenge_poly_eval) == ModuloCircuitElement
        assert type(root_power_times_tp_gemini_r) == ModuloCircuitElement

        circuit.extend_struct_output(
            u384("challenge_poly_eval", elmts=[challenge_poly_eval])
        )
        circuit.extend_struct_output(
            u384("root_power_times_tp_gemini_r", elmts=[root_power_times_tp_gemini_r])
        )
        return circuit


class ZKEvalsConsistencyLoopCircuit(ZKBaseUltraHonkCircuit):
    def __init__(
        self,
        vk: HonkVk,
        curve_id: int = CurveID.GRUMPKIN.value,
        auto_run: bool = True,
        compilation_mode: int = 1,
    ) -> None:
        name = f"zk_honk_evals_cons_loop_size_{vk.log_circuit_size}"
        self.vk = vk
        super().__init__(
            name, vk.log_circuit_size, curve_id, auto_run, compilation_mode
        )

    @property
    def input_map(self) -> dict:
        imap = {}

        imap["challenge_poly_eval"] = structs.u384
        imap["root_power_times_tp_gemini_r"] = structs.u384
        imap["tp_sumcheck_u_challenge"] = structs.u384

        return imap

    def _execute_circuit_logic(
        self, circuit: ZKHonkVerifierCircuits, vars: dict
    ) -> ModuloCircuit:

        challenge_poly_eval, root_power_times_tp_gemini_r = (
            circuit._check_evals_consistency_loop(
                vars["challenge_poly_eval"],
                vars["root_power_times_tp_gemini_r"],
                vars["tp_sumcheck_u_challenge"],
            )
        )

        assert type(challenge_poly_eval) == ModuloCircuitElement
        assert type(root_power_times_tp_gemini_r) == ModuloCircuitElement

        circuit.extend_struct_output(
            u384("challenge_poly_eval", elmts=[challenge_poly_eval])
        )
        circuit.extend_struct_output(
            u384("root_power_times_tp_gemini_r", elmts=[root_power_times_tp_gemini_r])
        )
        return circuit


class ZKEvalsConsistencyDoneCircuit(ZKBaseUltraHonkCircuit):
    def __init__(
        self,
        vk: HonkVk,
        curve_id: int = CurveID.GRUMPKIN.value,
        auto_run: bool = True,
        compilation_mode: int = 1,
    ) -> None:
        name = f"zk_honk_evals_cons_done_size_{vk.log_circuit_size}"
        self.vk = vk
        super().__init__(
            name, vk.log_circuit_size, curve_id, auto_run, compilation_mode
        )

    @property
    def input_map(self) -> dict:
        imap = {}

        imap["p_libra_evaluation"] = structs.u384
        imap["p_libra_poly_evals"] = (structs.u256Span, 4)
        imap["tp_gemini_r"] = structs.u384
        imap["challenge_poly_eval"] = structs.u384
        imap["root_power_times_tp_gemini_r"] = structs.u384

        return imap

    def _execute_circuit_logic(
        self, circuit: ZKHonkVerifierCircuits, vars: dict
    ) -> ModuloCircuit:

        vanishing_check, diff_check = circuit._check_evals_consistency_done(
            vars["p_libra_evaluation"],
            vars["p_libra_poly_evals"],
            vars["tp_gemini_r"],
            vars["challenge_poly_eval"],
            vars["root_power_times_tp_gemini_r"],
        )

        assert type(vanishing_check) == ModuloCircuitElement
        assert type(diff_check) == ModuloCircuitElement

        circuit.extend_struct_output(u384("vanishing_check", elmts=[vanishing_check]))
        circuit.extend_struct_output(u384("diff_check", elmts=[diff_check]))
        return circuit



from garaga.algebra import ModuloCircuitElement
from garaga.definitions import CurveID
from garaga.precompiled_circuits.ec import BasicEC, BasicECG2


class SlowG2CofactorClearing(BasicECG2):
    """
    Implements cofactor clearing for G2 points on BLS12-381 curve.
    Based on Teku's implementation of BLS signatures.
    """

    def h2_chain(
        self, P: tuple[list[ModuloCircuitElement], list[ModuloCircuitElement]]
    ):
        """
        Addition chain for multiplication by the E2 cofactor.

        Reference implementation:
        https://github.com/Consensys/teku/blob/55d04f87b422112312f79c1b4d662b3d58e3ca74/bls/src/main/java/tech/pegasys/teku/bls/impl/mikuli/hash2g2/Chains.java#L569
        """
        assert (
            self.curve_id == CurveID.BLS12_381.value
        ), "This circuit is only supported for BLS12-381"

        # Initial point is t0 = P
        t0 = P

        t1 = self.double_point_a_eq_0(t0)
        t4 = self.add_points(t1, t0)
        t2 = self.add_points(t4, t1)
        t3 = self.add_points(t2, t1)
        t11 = self.add_points(t3, t1)
        t9 = self.add_points(t11, t1)
        t10 = self.add_points(t9, t1)
        t5 = self.add_points(t10, t1)
        t7 = self.add_points(t5, t1)
        t15 = self.add_points(t7, t1)
        t13 = self.add_points(t15, t1)
        t6 = self.add_points(t13, t1)
        t14 = self.add_points(t6, t1)
        t12 = self.add_points(t14, t1)
        t8 = self.add_points(t12, t1)

        # Start the main computation chain
        t1 = self.double_point_a_eq_0(t6)  # t6.dbl()
        t1 = self.double_n_times(t1, 5)  # .dbls(5)
        t1 = self.add_points(t1, t13)  # .add(t13)
        t1 = self.double_n_times(t1, 2)  # .dbls(2)
        t1 = self.add_points(t1, t0)  # .add(t0)
        t1 = self.double_n_times(t1, 9)  # .dbls(9)
        t1 = self.add_points(t1, t8)  # .add(t8)
        t1 = self.double_n_times(t1, 5)  # .dbls(5)
        t1 = self.add_points(t1, t11)  # .add(t11)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t13)  # .add(t13)
        t1 = self.double_n_times(t1, 8)  # .dbls(8)
        t1 = self.add_points(t1, t2)  # .add(t2)
        t1 = self.double_n_times(t1, 5)  # .dbls(5)
        t1 = self.add_points(t1, t3)  # .add(t3)
        t1 = self.double_n_times(t1, 5)  # .dbls(5)
        t1 = self.add_points(t1, t3)  # .add(t3)
        t1 = self.double_n_times(t1, 4)  # .dbls(4)
        t1 = self.add_points(t1, t5)  # .add(t5)
        t1 = self.double_n_times(t1, 4)  # .dbls(4)
        t1 = self.add_points(t1, t0)  # .add(t0)
        t1 = self.double_n_times(t1, 8)  # .dbls(8)
        t1 = self.add_points(t1, t11)  # .add(t11)
        t1 = self.double_n_times(t1, 8)  # .dbls(8)
        t1 = self.add_points(t1, t8)  # .add(t8)
        t1 = self.double_n_times(t1, 4)  # .dbls(4)
        t1 = self.add_points(t1, t2)  # .add(t2)
        t1 = self.double_n_times(t1, 9)  # .dbls(9)
        t1 = self.add_points(t1, t5)  # .add(t5)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t11)  # .add(t11)
        t1 = self.double_n_times(t1, 2)  # .dbls(2)
        t1 = self.add_points(t1, t0)  # .add(t0)
        t1 = self.double_n_times(t1, 9)  # .dbls(9)
        t1 = self.add_points(t1, t8)  # .add(t8)
        t1 = self.double_n_times(t1, 5)  # .dbls(5)
        t1 = self.add_points(t1, t13)  # .add(t13)
        t1 = self.double_n_times(t1, 4)  # .dbls(4)
        t1 = self.add_points(t1, t0)  # .add(t0)
        t1 = self.double_n_times(t1, 11)  # .dbls(11)
        t1 = self.add_points(t1, t9)  # .add(t9)
        t1 = self.double_n_times(t1, 7)  # .dbls(7)
        t1 = self.add_points(t1, t12)  # .add(t12)
        t1 = self.double_n_times(t1, 7)  # .dbls(7)
        t1 = self.add_points(t1, t7)  # .add(t7)
        t1 = self.double_n_times(t1, 5)  # .dbls(5)
        t1 = self.add_points(t1, t12)  # .add(t12)
        t1 = self.double_n_times(t1, 5)  # .dbls(5)
        t1 = self.add_points(t1, t14)  # .add(t14)
        t1 = self.double_n_times(t1, 8)  # .dbls(8)
        t1 = self.add_points(t1, t13)  # .add(t13)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t3)  # .add(t3)
        t1 = self.double_n_times(t1, 5)  # .dbls(5)
        t1 = self.add_points(t1, t0)  # .add(t0)
        t1 = self.double_n_times(t1, 8)  # .dbls(8)
        t1 = self.add_points(t1, t9)  # .add(t9)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t13)  # .add(t13)
        t1 = self.double_n_times(t1, 4)  # .dbls(4)
        t1 = self.add_points(t1, t10)  # .add(t10)
        t1 = self.double_n_times(t1, 4)  # .dbls(4)
        t1 = self.add_points(t1, t2)  # .add(t2)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t10)  # .add(t10)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t2)  # .add(t2)
        t1 = self.double_n_times(t1, 4)  # .dbls(4)
        t1 = self.add_points(t1, t0)  # .add(t0)
        t1 = self.double_n_times(t1, 10)  # .dbls(10)
        t1 = self.add_points(t1, t9)  # .add(t9)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t14)  # .add(t14)
        t1 = self.double_n_times(t1, 4)  # .dbls(4)
        t1 = self.add_points(t1, t3)  # .add(t3)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t9)  # .add(t9)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t15)  # .add(t15)
        t1 = self.double_n_times(t1, 5)  # .dbls(5)
        t1 = self.add_points(t1, t8)  # .add(t8)
        t1 = self.double_n_times(t1, 5)  # .dbls(5)
        t1 = self.add_points(t1, t12)  # .add(t12)
        t1 = self.double_n_times(t1, 4)  # .dbls(4)
        t1 = self.add_points(t1, t5)  # .add(t5)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t15)  # .add(t15)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t2)  # .add(t2)
        t1 = self.double_n_times(t1, 7)  # .dbls(7)
        t1 = self.add_points(t1, t5)  # .add(t5)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t3)  # .add(t3)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t9)  # .add(t9)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t15)  # .add(t15)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t14)  # .add(t14)
        t1 = self.double_n_times(t1, 5)  # .dbls(5)
        t1 = self.add_points(t1, t8)  # .add(t8)
        t1 = self.double_n_times(t1, 10)  # .dbls(10)
        t1 = self.add_points(t1, t6)  # .add(t6)
        t1 = self.double_n_times(t1, 5)  # .dbls(5)
        t1 = self.add_points(t1, t5)  # .add(t5)
        t1 = self.double_n_times(t1, 3)  # .dbls(3)
        t1 = self.add_points(t1, t0)  # .add(t0)
        t1 = self.double_n_times(t1, 9)  # .dbls(9)
        t1 = self.add_points(t1, t13)  # .add(t13)
        t1 = self.double_n_times(t1, 7)  # .dbls(7)
        t1 = self.add_points(t1, t12)  # .add(t12)
        t1 = self.double_n_times(t1, 4)  # .dbls(4)
        t1 = self.add_points(t1, t5)  # .add(t5)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t2)  # .add(t2)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t11)  # .add(t11)
        t1 = self.double_n_times(t1, 4)  # .dbls(4)
        t1 = self.add_points(t1, t10)  # .add(t10)
        t1 = self.double_n_times(t1, 4)  # .dbls(4)
        t1 = self.add_points(t1, t4)  # .add(t4)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t10)  # .add(t10)
        t1 = self.double_n_times(t1, 7)  # .dbls(7)
        t1 = self.add_points(t1, t7)  # .add(t7)
        t1 = self.double_n_times(t1, 3)  # .dbls(3)
        t1 = self.add_points(t1, t2)  # .add(t2)
        t1 = self.double_n_times(t1, 4)  # .dbls(4)
        t1 = self.add_points(t1, t3)  # .add(t3)
        t1 = self.double_n_times(t1, 8)  # .dbls(8)
        t1 = self.add_points(t1, t9)  # .add(t9)
        t1 = self.double_n_times(t1, 8)  # .dbls(8)
        t1 = self.add_points(t1, t9)  # .add(t9)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t8)  # .add(t8)
        t1 = self.double_n_times(t1, 5)  # .dbls(5)
        t1 = self.add_points(t1, t7)  # .add(t7)
        t1 = self.double_n_times(t1, 5)  # .dbls(5)
        t1 = self.add_points(t1, t6)  # .add(t6)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t5)  # .add(t5)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t4)  # .add(t4)
        t1 = self.double_n_times(t1, 5)  # .dbls(5)
        t1 = self.add_points(t1, t5)  # .add(t5)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t4)  # .add(t4)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t3)  # .add(t3)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t4)  # .add(t4)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t5)  # .add(t5)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t3)  # .add(t3)
        t1 = self.double_n_times(t1, 7)  # .dbls(7)
        t1 = self.add_points(t1, t3)  # .add(t3)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t3)  # .add(t3)
        t1 = self.double_n_times(t1, 5)  # .dbls(5)
        t1 = self.add_points(t1, t4)  # .add(t4)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t3)  # .add(t3)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t3)  # .add(t3)
        t1 = self.double_n_times(t1, 3)  # .dbls(3)
        t1 = self.add_points(t1, t0)  # .add(t0)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t3)  # .add(t3)
        t1 = self.double_n_times(t1, 6)  # .dbls(6)
        t1 = self.add_points(t1, t3)  # .add(t3)
        t1 = self.double_n_times(t1, 5)  # .dbls(5)
        t1 = self.add_points(t1, t2)  # .add(t2)

        return t1

    def mx_chain(self, P):
        """
        Addition chain for multiplication by 0xd201000000010000 == -x,
        where x is the BLS parameter.
        """
        q = self.double_point_a_eq_0(P)

        # Sequence of doubles and adds as per Java implementation
        ndoubles = [2, 3, 9, 32, 16]
        for n in ndoubles:
            q = self.add_points(q, P)
            for _ in range(n):
                q = self.double_point_a_eq_0(q)

        return q

    def clear_cofactor(self, P):
        """
        Complete cofactor clearing implementation following Teku's approach.

        This implements an optimized cofactor clearing method for BLS12-381 G2 that avoids
        patent-encumbered endomorphism techniques. Instead of directly multiplying by the G2
        cofactor, it uses an efficient addition chain to compute an equivalent operation.

        The method follows the approach specified in draft-irtf-cfrg-hash-to-curve-05 section 8.9.2,
        which breaks down the operation into:
        1. Apply h2 chain
        2. Multiply by 3
        3. Apply mx chain twice (z^2)
        4. Subtract 3 * h2 result

        This gives the same result as multiplying by the effective cofactor, but is more
        computationally efficient due to the optimized addition chain implementation.

        Args:
            P: The point to clear the cofactor from

        Returns:
            A point in the correct subgroup after cofactor clearing
        """
        assert (
            self.curve_id == CurveID.BLS12_381.value
        ), "This circuit is only supported for BLS12-381"

        # Step 1: Apply h2 chain
        work = self.h2_chain(P)

        # Step 2: Multiply by 3 (add + double)
        work3 = self.add_points(work, self.double_point_a_eq_0(work))

        # Step 3: Apply mx chain twice (z^2)
        work = self.mx_chain(work3)  # First z
        work = self.mx_chain(work)  # Second z

        # Step 4: Subtract 3 * h2
        neg_work3 = self.negate_point(work3)
        final = self.add_points(work, neg_work3)

        return final


class FastG2CofactorClearing(BasicECG2):
    """
    Implements an efficient cofactor clearing method for G2 points on BLS12-381 curve
    using endomorphisms. This implementation avoids expensive scalar multiplications
    by utilizing the Frobenius endomorphism (ψ) and efficient addition chains.

    The cofactor clearing ensures that points are in the correct order-r subgroup of G2,
    which is essential for cryptographic operations like BLS signatures.

    Key operations:
    - ψ (psi): Frobenius endomorphism
    - ψ² (psi²): Double application of Frobenius
    - mul_by_x: Multiplication by BLS parameter x
    """

    def psi(
        self, P: tuple[list[ModuloCircuitElement], list[ModuloCircuitElement]]
    ) -> tuple[list[ModuloCircuitElement], list[ModuloCircuitElement]]:
        """
        Implements the ψ (Frobenius endomorphism) map for BLS12-381 G₂ points.

        The Frobenius endomorphism ψ is defined as:
        ψ(P) = π(P) · [1/(u+1)^((p-1)/3), 1/(u+1)^((p-1)/2)]
        where π is the p-power Frobenius map that conjugates coordinates in Fp2.

        Args:
            P: A G2 point represented as ((x0, x1), (y0, y1)) where each coordinate
               is in Fp and (x0, x1) and (y0, y1) represent elements in Fp2

        Returns:
            A new G2 point after applying the Frobenius endomorphism
        """
        assert (
            self.curve_id == CurveID.BLS12_381.value
        ), "This circuit is only supported for BLS12-381"

        x, y = P  # x and y are lists [x0, x1] representing elements in Fp²

        # Apply the Frobenius map (conjugation in Fp2)
        x_frob = self.fp2_frobenius_map(x)
        y_frob = self.fp2_frobenius_map(y)

        # Constants for scaling after Frobenius map
        # PSI_X = 1/(u+1)^((p-1)/3)
        psi_coeff_x = [
            self.set_or_get_constant(self.field(0)),
            self.set_or_get_constant(
                self.field(
                    0x1A0111EA397FE699EC02408663D4DE85AA0D857D89759AD4897D29650FB85F9B409427EB4F49FFFD8BFD00000000AAAD
                )
            ),
        ]

        # PSI_Y = 1/(u+1)^((p-1)/2)
        psi_coeff_y = [
            self.set_or_get_constant(
                self.field(
                    0x135203E60180A68EE2E9C448D77A2CD91C3DEDD930B1CF60EF396489F61EB45E304466CF3E67FA0AF1EE7B04121BDEA2
                )
            ),
            self.set_or_get_constant(
                self.field(
                    0x6AF0E0437FF400B6831E36D6BD17FFE48395DABC2D3435E77F76E17009241C5EE67992F72EC05F4C81084FBEDE3CC09
                )
            ),
        ]

        # Scale the coordinates with the appropriate constants
        psi_x = self.fp2_mul(x_frob, psi_coeff_x)
        psi_y = self.fp2_mul(y_frob, psi_coeff_y)

        return (psi_x, psi_y)

    def mul_by_x(
        self, P: tuple[list[ModuloCircuitElement], list[ModuloCircuitElement]]
    ) -> tuple[list[ModuloCircuitElement], list[ModuloCircuitElement]]:
        """
        Multiplies a point by x = -0xd201000000010001 (BLS parameter) using double-and-add.

        The binary representation of x is:
        1101 0010 0000 0001 0000 0000 0000 0000 0000 0000 0000 0001 0000 0000 0000 0000

        Args:
            P: A G2 point to be multiplied by x

        Returns:
            [-x]P: The point multiplied by the negation of x
        """
        assert (
            self.curve_id == CurveID.BLS12_381.value
        ), "This circuit is only supported for BLS12-381"

        result = P  # 1: Start with P. As a result we skip the first bit.
        result = self.double_point_a_eq_0(result)  # 1: Double and add
        result = self.add_points(result, P)  # 2: Double and add
        result = self.double_point_a_eq_0(result)  # 0: Double
        result = self.double_point_a_eq_0(result)  # 1: Double and add
        result = self.add_points(result, P)
        result = self.double_n_times(result, 2)  # 2x 0: Double
        result = self.double_point_a_eq_0(result)  # 1: Double and add
        result = self.add_points(result, P)
        result = self.double_n_times(result, 8)  # 8x 0: Double
        result = self.double_point_a_eq_0(result)  # 1: Double and add
        result = self.add_points(result, P)
        result = self.double_n_times(result, 31)  # 31x 0: Double
        result = self.double_point_a_eq_0(result)  # 1: Double and add
        result = self.add_points(result, P)
        result = self.double_n_times(result, 16)  # 16x 0: Double

        return self.negate_point(result)

    def double_psi(
        self, P: tuple[list[ModuloCircuitElement], list[ModuloCircuitElement]]
    ) -> tuple[list[ModuloCircuitElement], list[ModuloCircuitElement]]:
        """
        Implements ψ² (double Frobenius endomorphism) by applying psi twice.

        For BLS12-381, ψ² is a much cheaper operation than scalar multiplication
        and is used to optimize the cofactor clearing operation.

        Args:
            P: A G2 point

        Returns:
            ψ²(P): The point after applying the Frobenius endomorphism twice
        """
        return self.psi(self.psi(P))

    def clear_cofactor(
        self, P: tuple[list[ModuloCircuitElement], list[ModuloCircuitElement]]
    ) -> tuple[list[ModuloCircuitElement], list[ModuloCircuitElement]]:
        """
        Clears the cofactor of a G2 point using an optimized endomorphism-based method.

        This implementation follows the efficient approach described in:
        https://eprint.iacr.org/2017/419.pdf

        The formula implemented is:
        [h(ψ)]P = [x²-x-1]P + [x-1]ψ(P) + ψ²(2P)
        where x is the BLS parameter = -0xd201000000010000

        This method is more efficient than the traditional scalar multiplication
        by the cofactor h, as it uses the efficiently computable endomorphism ψ
        and optimized multiplication by x.

        Args:
            P: A G2 point to clear the cofactor from

        Returns:
            A point in the correct order-r subgroup of G2
        """
        # Calculate [-x]P
        t1 = self.mul_by_x(P)

        # Calculate ψ(P)
        t2 = self.psi(P)

        # Calculate 2P and ψ²(2P)
        t3 = self.double_point_a_eq_0(P)  # 2P
        t3 = self.double_psi(t3)  # ψ²(2P)

        # Following steps compute: [x²-x-1]P + [x-1]ψ(P) + ψ²(2P)
        t3 = self.add_points(t3, self.negate_point(t2))  # ψ²(2P) - ψ(P)
        t2 = self.add_points(t1, t2)  # [-x]P + ψ(P)
        t2 = self.mul_by_x(t2)  # [x²]P - [x]ψ(P)
        t3 = self.add_points(t3, t2)  # ψ²(2P) - ψ(P) + [x²]P - [x]ψ(P)
        t3 = self.add_points(t3, self.negate_point(t1))  # Add [x]P
        final = self.add_points(t3, self.negate_point(P))  # Subtract P

        return final


class G1CofactorClearing(BasicEC):

    def mul_by_x(
        self, P: tuple[ModuloCircuitElement, ModuloCircuitElement]
    ) -> tuple[ModuloCircuitElement, ModuloCircuitElement]:
        """
        Multiplies a point by x = 0xd201000000010000 (BLS parameter) using double-and-add.

        The binary representation of x is:
        1101 0010 0000 0001 0000 0000 0000 0000 0000 0000 0000 0001 0000 0000 0000 0000

        Args:
            P: A G1 point to be multiplied by x

        Returns:
            [x]P: The point multiplied by x
        """

        # Todo : Incude automatic addition chain for any number computed at runtime ?
        assert (
            self.curve_id == CurveID.BLS12_381.value
        ), "This circuit is only supported for BLS12-381"

        result = P  # 1: Start with P. As a result we skip the first bit.
        result = self.double_point_a_eq_0(result)  # 1: Double and add
        result = self.add_points(result, P)  # 2: Double and add
        result = self.double_point_a_eq_0(result)  # 0: Double
        result = self.double_point_a_eq_0(result)  # 1: Double and add
        result = self.add_points(result, P)
        result = self.double_n_times(result, 2)  # 2x 0: Double
        result = self.double_point_a_eq_0(result)  # 1: Double and add
        result = self.add_points(result, P)
        result = self.double_n_times(result, 8)  # 8x 0: Double
        result = self.double_point_a_eq_0(result)  # 1: Double and add
        result = self.add_points(result, P)
        result = self.double_n_times(result, 31)  # 31x 0: Double
        result = self.double_point_a_eq_0(result)  # 1: Double and add
        result = self.add_points(result, P)
        return self.double_n_times(result, 16)  # 16x 0: Double

    def clear_cofactor(
        self, P: tuple[ModuloCircuitElement, ModuloCircuitElement]
    ) -> tuple[ModuloCircuitElement, ModuloCircuitElement]:
        """
        Clears the cofactor of a G1 point using an efficient scalar multiplication method.

        Implementation follows the approach from:
        https://eprint.iacr.org/2019/403

        The formula is: h_eff(P) = P - [-x]P = P + [x]P
        where x is the BLS parameter = -0xd201000000010000

        This method is more efficient than multiplying by the full cofactor h,
        as it uses an optimized addition chain for multiplication by x.

        Args:
            P: A G1 point to clear the cofactor from

        Returns:
            A point in the correct order-r subgroup of G1
        """
        assert (
            self.curve_id == CurveID.BLS12_381.value
        ), "This circuit is only supported for BLS12-381"

        # Calculate [x]P using the existing mul_by_x method
        # As shown below, we can save negation
        xP = self.mul_by_x(P)

        # Return P - [-x]P = P + [x]P
        return self.add_points(P, xP)


if __name__ == "__main__":
    circuit = G1CofactorClearing(name="cofactor_clearing", curve_id=1)
    val_x = circuit.write_element(
        circuit.field(
            3789617024712504402204306620295003375951143917889162928515122476381982967144814366712031831841518399614182231387665
        )
    )

    val_y = circuit.write_element(
        circuit.field(
            1292375129422168617658520396283100687686347104559592203462491249161639006037671760603453326853098986903549775136448
        )
    )

    print(circuit.clear_cofactor((val_x, val_y)))
